"""Miscellaneous utilities"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import asyncio
import concurrent.futures
import errno
import hashlib
import inspect
import random
import secrets
import socket
import ssl
import sys
import uuid
import warnings
from binascii import b2a_hex
from datetime import datetime, timezone
from hmac import compare_digest
from operator import itemgetter
from typing import Any, Dict, List
from urllib.parse import quote

from async_generator import aclosing
from tornado import gen, ioloop, web
from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado.log import app_log


def random_port():
    """Get a single random port."""
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


# ISO8601 for strptime with/without milliseconds
ISO8601_ms = "%Y-%m-%dT%H:%M:%S.%fZ"
ISO8601_s = "%Y-%m-%dT%H:%M:%SZ"


def isoformat(dt):
    """Render a datetime object as an ISO 8601 UTC timestamp

    Naive datetime objects are assumed to be UTC
    """
    # allow null timestamps to remain None without
    # having to check if isoformat should be called
    if dt is None:
        return None
    if dt.tzinfo:
        dt = dt.astimezone(timezone.utc).replace(tzinfo=None)
    return dt.isoformat() + "Z"


def can_connect(ip, port):
    """Check if we can connect to an ip:port.

    Return True if we can connect, False otherwise.
    """
    if ip in {"", "0.0.0.0", "::"}:
        ip = "127.0.0.1"
    try:
        socket.create_connection((ip, port)).close()
    except OSError as e:
        if e.errno not in {errno.ECONNREFUSED, errno.ETIMEDOUT}:
            app_log.error("Unexpected error connecting to %s:%i %s", ip, port, e)
        return False
    else:
        return True


def make_ssl_context(
    keyfile,
    certfile,
    cafile=None,
    verify=None,
    check_hostname=None,
    purpose=ssl.Purpose.SERVER_AUTH,
):
    """Setup context for starting an https server or making requests over ssl.

    Used for verifying internal ssl connections.
    Certificates are always verified in both directions.
    Hostnames are checked for client sockets.

    Client sockets are created with `purpose=ssl.Purpose.SERVER_AUTH` (default),
    Server sockets are created with `purpose=ssl.Purpose.CLIENT_AUTH`.
    """
    if not keyfile or not certfile:
        return None
    if verify is not None:
        purpose = ssl.Purpose.SERVER_AUTH if verify else ssl.Purpose.CLIENT_AUTH
        warnings.warn(
            f"make_ssl_context(verify={verify}) is deprecated in jupyterhub 2.4."
            f" Use make_ssl_context(purpose={purpose!s}).",
            DeprecationWarning,
            stacklevel=2,
        )
    if check_hostname is not None:
        purpose = ssl.Purpose.SERVER_AUTH if check_hostname else ssl.Purpose.CLIENT_AUTH
        warnings.warn(
            f"make_ssl_context(check_hostname={check_hostname}) is deprecated in jupyterhub 2.4."
            f" Use make_ssl_context(purpose={purpose!s}).",
            DeprecationWarning,
            stacklevel=2,
        )

    ssl_context = ssl.create_default_context(purpose, cafile=cafile)
    # always verify
    ssl_context.verify_mode = ssl.CERT_REQUIRED

    if purpose == ssl.Purpose.SERVER_AUTH:
        # SERVER_AUTH is authenticating servers (i.e. for a client)
        ssl_context.check_hostname = True
    ssl_context.load_default_certs()

    ssl_context.load_cert_chain(certfile, keyfile)
    ssl_context.check_hostname = check_hostname
    return ssl_context


# AnyTimeoutError catches TimeoutErrors coming from asyncio, tornado, stdlib
AnyTimeoutError = (gen.TimeoutError, asyncio.TimeoutError, TimeoutError)


async def exponential_backoff(
    pass_func,
    fail_message,
    start_wait=0.2,
    scale_factor=2,
    max_wait=5,
    timeout=10,
    timeout_tolerance=0.1,
    *args,
    **kwargs,
):
    """
    Exponentially backoff until `pass_func` is true.

    The `pass_func` function will wait with **exponential backoff** and
    **random jitter** for as many needed iterations of the Tornado loop,
    until reaching maximum `timeout` or truthiness. If `pass_func` is still
    returning false at `timeout`, a `TimeoutError` will be raised.

    The first iteration will begin with a wait time of `start_wait` seconds.
    Each subsequent iteration's wait time will scale up by continuously
    multiplying itself by `scale_factor`. This continues for each iteration
    until `pass_func` returns true or an iteration's wait time has reached
    the `max_wait` seconds per iteration.

    `pass_func` may be a future, although that is not entirely recommended.

    Parameters
    ----------
    pass_func
        function that is to be run
    fail_message : str
        message for a `TimeoutError`
    start_wait : optional
        initial wait time for the first iteration in seconds
    scale_factor : optional
        a multiplier to increase the wait time for each iteration
    max_wait : optional
        maximum wait time per iteration in seconds
    timeout : optional
        maximum time of total wait in seconds
    timeout_tolerance : optional
        a small multiplier used to add jitter to `timeout`'s deadline
    *args, **kwargs
        passed to `pass_func(*args, **kwargs)`

    Returns
    -------
    value of `pass_func(*args, **kwargs)`

    Raises
    ------
    TimeoutError
        If `pass_func` is still false at the end of the `timeout` period.

    Notes
    -----
    See https://www.awsarchitectureblog.com/2015/03/backoff.html
    for information about the algorithm and examples. We're using their
    full Jitter implementation equivalent.
    """
    loop = ioloop.IOLoop.current()
    deadline = loop.time() + timeout
    # add jitter to the deadline itself to prevent re-align of a bunch of
    # timing out calls once the deadline is reached.
    if timeout_tolerance:
        tol = timeout_tolerance * timeout
        deadline = random.uniform(deadline - tol, deadline + tol)
    scale = 1
    while True:
        ret = await maybe_future(pass_func(*args, **kwargs))
        # Truthy!
        if ret:
            return ret
        remaining = deadline - loop.time()
        if remaining < 0:
            # timeout exceeded
            break
        # add some random jitter to improve performance
        # this prevents overloading any single tornado loop iteration with
        # too many things
        limit = min(max_wait, start_wait * scale)
        if limit < max_wait:
            scale *= scale_factor
        dt = min(remaining, random.uniform(0, limit))
        await asyncio.sleep(dt)
    raise asyncio.TimeoutError(fail_message)


async def wait_for_server(ip, port, timeout=10):
    """Wait for any server to show up at ip:port."""
    if ip in {"", "0.0.0.0", "::"}:
        ip = "127.0.0.1"
    await exponential_backoff(
        lambda: can_connect(ip, port),
        "Server at {ip}:{port} didn't respond in {timeout} seconds".format(
            ip=ip, port=port, timeout=timeout
        ),
        timeout=timeout,
    )


async def wait_for_http_server(url, timeout=10, ssl_context=None):
    """Wait for an HTTP Server to respond at url.

    Any non-5XX response code will do, even 404.
    """
    ioloop.IOLoop.current()
    client = AsyncHTTPClient()
    if ssl_context:
        client.ssl_options = ssl_context

    async def is_reachable():
        try:
            r = await client.fetch(url, follow_redirects=False)
            return r
        except HTTPError as e:
            if e.code >= 500:
                # failed to respond properly, wait and try again
                if e.code != 599:
                    # we expect 599 for no connection,
                    # but 502 or other proxy error is conceivable
                    app_log.warning("Server at %s responded with error: %s", url, e.code)
            else:
                app_log.debug("Server at %s responded with %s", url, e.code)
                return e.response
        except OSError as e:
            if e.errno not in {errno.ECONNABORTED, errno.ECONNREFUSED, errno.ECONNRESET}:
                app_log.warning("Failed to connect to %s (%s)", url, e)
        return False

    re = await exponential_backoff(
        is_reachable,
        "Server at {url} didn't respond in {timeout} seconds".format(url=url, timeout=timeout),
        timeout=timeout,
    )
    return re


# Decorators for authenticated Handlers
def auth_decorator(check_auth):
    """Make an authentication decorator.

    I heard you like decorators, so I put a decorator
    in your decorator, so you can decorate while you decorate.
    """

    def decorator(method):
        def decorated(self, *args, **kwargs):
            check_auth(self, **kwargs)
            return method(self, *args, **kwargs)

        # Perhaps replace with functools.wrap
        decorated.__name__ = method.__name__
        decorated.__doc__ = method.__doc__
        return decorated

    decorator.__name__ = check_auth.__name__
    decorator.__doc__ = check_auth.__doc__
    return decorator


@auth_decorator
def token_authenticated(self):
    """Decorator for method authenticated only by Authorization token header

    (no cookies)
    """
    if self.get_current_user_token() is None:
        raise web.HTTPError(403)


@auth_decorator
def authenticated_403(self):
    """Decorator for method to raise 403 error instead of redirect to login

    Like tornado.web.authenticated, this decorator raises a 403 error
    instead of redirecting to login.
    """
    if self.current_user is None:
        raise web.HTTPError(403)


def admin_only(f):
    """Deprecated!"""
    # write it this way to trigger deprecation warning at decoration time,
    # not on the method call
    warnings.warn(
        """@jupyterhub.utils.admin_only is deprecated in JupyterHub 2.0.

        Use the new `@jupyterhub.scopes.needs_scope` decorator to resolve permissions,
        or check against `self.current_user.parsed_scopes`.
        """,
        DeprecationWarning,
        stacklevel=2,
    )

    # the original decorator
    @auth_decorator
    def admin_only(self):
        """Decorator for restricting access to admin users"""
        user = self.current_user
        if user is None or not user.admin:
            raise web.HTTPError(403)

    return admin_only(f)


@auth_decorator
def metrics_authentication(self):
    """Decorator for restricting access to metrics"""
    if not self.authenticate_prometheus:
        return
    scope = "read:metrics"
    if scope not in self.parsed_scopes:
        raise web.HTTPError(403, reason=f"Access to metrics requires scope '{scope}'")


# Token utilities


def new_token(*args, **kwargs):
    """Generator for new random tokens

    For now, just UUIDs.
    """
    return uuid.uuid4().hex


def hash_token(token, salt=8, rounds=16384, algorithm="sha512"):
    """Hash a token, and return it as `algorithm:salt:hash`.

    If `salt` is an integer, a random salt of that many bytes will be used.
    """
    h = hashlib.new(algorithm)
    if isinstance(salt, int):
        salt = b2a_hex(secrets.token_bytes(salt))
    if isinstance(salt, bytes):
        bsalt = salt
        salt = salt.decode("utf8")
    else:
        bsalt = salt.encode("utf8")
    btoken = token.encode("utf8", "replace")
    h.update(bsalt)
    for i in range(rounds):
        h.update(btoken)
    digest = h.hexdigest()

    return f"{algorithm}:{rounds}:{salt}:{digest}"


def compare_token(compare, token):
    """Compare a token with a hashed token.

    Uses the same algorithm and salt of the hashed token for comparison.
    """
    algorithm, srounds, salt, _ = compare.split(":")
    hashed = hash_token(token, salt=salt, rounds=int(srounds), algorithm=algorithm).encode("utf8")
    compare = compare.encode("utf8")
    if compare_digest(compare, hashed):
        return True
    return False


def url_escape_path(value):
    """Escape a value to be used in URLs, cookies, etc."""
    return quote(value, safe="@~")


def url_path_join(*pieces):
    """Join components of url into a relative url.

    Use to prevent double slash when joining subpath. This will leave the
    initial and final / in place.

    Copied from `notebook.utils.url_path_join`.
    """
    initial = pieces[0].startswith("/")
    final = pieces[-1].endswith("/")
    stripped = [s.strip("/") for s in pieces]
    result = "/".join(s for s in stripped if s)

    if initial:
        result = "/" + result
    if final:
        result = result + "/"
    if result == "//":
        result = "/"

    return result


def print_ps_info(file=sys.stderr):
    """Print process summary info from psutil

    warns if psutil is unavailable
    """
    try:
        import psutil
    except ImportError:
        # nothing to print
        warnings.warn(
            "psutil unavailable. Install psutil to get CPU and memory stats", stacklevel=2
        )
        return
    p = psutil.Process()
    # format CPU percentage
    cpu = p.cpu_percent(0.1)
    if cpu >= 10:
        cpu_s = "%i" % cpu
    else:
        cpu_s = "%.1f" % cpu

    # format memory (only resident set)
    rss = p.memory_info().rss
    if rss >= 1e9:
        mem_s = "%.1fG" % (rss / 1e9)
    elif rss >= 1e7:
        mem_s = "%.0fM" % (rss / 1e6)
    elif rss >= 1e6:
        mem_s = "%.1fM" % (rss / 1e6)
    else:
        mem_s = "%.0fk" % (rss / 1e3)

    # left-justify and shrink-to-fit columns
    cpulen = max(len(cpu_s), 4)
    memlen = max(len(mem_s), 3)
    fd_s = str(p.num_fds())
    fdlen = max(len(fd_s), 3)

    print(
        "%s %s %s %s" % ("%CPU".ljust(cpulen), "MEM".ljust(memlen), "FDs".ljust(fdlen), "threads"),
        file=file,
    )

    print(
        "%s %s %s %s"
        % (
            cpu_s.ljust(cpulen),
            mem_s.ljust(memlen),
            fd_s.ljust(fdlen),
            str(p.num_threads()).ljust(7),
        ),
        file=file,
    )

    # trailing blank line
    print("", file=file)


def maybe_future(obj):
    """Return an asyncio Future

    Use instead of gen.maybe_future

    For our compatibility, this must accept:

    - asyncio coroutine (gen.maybe_future doesn't work in tornado < 5)
    - tornado coroutine (asyncio.ensure_future doesn't work)
    - scalar (asyncio.ensure_future doesn't work)
    - concurrent.futures.Future (asyncio.ensure_future doesn't work)
    - tornado Future (works both ways)
    - asyncio Future (works both ways)
    """
    if inspect.isawaitable(obj):
        # already awaitable, use ensure_future
        return asyncio.ensure_future(obj)
    elif isinstance(obj, concurrent.futures.Future):
        return asyncio.wrap_future(obj)
    else:
        # could also check for tornado.concurrent.Future
        # but with tornado >= 5.1 tornado.Future is asyncio.Future
        f = asyncio.Future()
        f.set_result(obj)
        return f


async def iterate_until(deadline_future, generator):
    """An async generator that yields items from a generator
    until a deadline future resolves

    This could *almost* be implemented as a context manager
    like asyncio_timeout with a Future for the cutoff.

    However, we want one distinction: continue yielding items
    after the future is complete, as long as the are already finished.

    Usage::

               async for item in iterate_until(some_future, some_async_generator()):
                   print(item)

    """
    async with aclosing(generator.__aiter__()) as aiter:
        while True:
            item_future = asyncio.ensure_future(aiter.__anext__())
            await asyncio.wait([item_future, deadline_future], return_when=asyncio.FIRST_COMPLETED)
            if item_future.done():
                try:
                    yield item_future.result()
                except (StopAsyncIteration, asyncio.CancelledError):
                    break
            elif deadline_future.done():
                # deadline is done *and* next item is not ready
                # cancel item future to avoid warnings about
                # unawaited tasks
                if not item_future.cancelled():
                    item_future.cancel()
                # resolve cancellation to avoid garbage collection issues
                try:
                    await item_future
                except asyncio.CancelledError:
                    pass
                break
            else:
                # neither is done, this shouldn't happen
                continue


def utcnow(*, with_tz=True):
    """Return utcnow

    with_tz (default): returns tz-aware datetime in UTC

    if with_tz=False, returns UTC timestamp without tzinfo
    (used for most internal timestamp storage because databases often don't preserve tz info)
    """
    now = datetime.now(timezone.utc)
    if not with_tz:
        now = now.replace(tzinfo=None)
    return now


def _parse_accept_header(accept):
    """
    Parse the Accept header *accept*

    Return a list with 3-tuples of
    [(str(media_type), dict(params), float(q_value)),] ordered by q values.
    If the accept header includes vendor-specific types like::
        application/vnd.yourcompany.yourproduct-v1.1+json
    It will actually convert the vendor and version into parameters and
    convert the content type into `application/json` so appropriate content
    negotiation decisions can be made.
    Default `q` for values that are not specified is 1.0

    From: https://gist.github.com/samuraisam/2714195
    """
    result = []
    for media_range in accept.split(","):
        parts = media_range.split(";")
        media_type = parts.pop(0).strip()
        media_params = []
        # convert vendor-specific content type to application/json
        typ, subtyp = media_type.split("/")
        # check for a + in the sub-type
        if "+" in subtyp:
            # if it exists, determine if the subtype is a vendor-specific type
            vnd, sep, extra = subtyp.partition("+")
            if vnd.startswith("vnd"):
                # and then... if it ends in something like "-v1.1" parse the
                # version out
                if "-v" in vnd:
                    vnd, sep, rest = vnd.rpartition("-v")
                    if rest:
                        # add the version as a media param
                        try:
                            version = float(rest)
                        except ValueError:
                            version = 1.0  # could not be parsed
                        media_params.append(("version", version))
                # add the vendor code as a media param
                media_params.append(("vendor", vnd))
                # and re-write media_type to something like application/json so
                # it can be used usefully when looking up emitters
                media_type = f"{typ}/{extra}"

        q = 1.0
        for part in parts:
            (key, value) = part.lstrip().split("=", 1)
            key = key.strip()
            value = value.strip()
            if key == "q":
                q = float(value)
            else:
                media_params.append((key, value))
        result.append((media_type, dict(media_params), q))
    result.sort(key=itemgetter(2))
    return result


def get_accepted_mimetype(accept_header, choices=None):
    """Return the preferred mimetype from an Accept header

    If `choices` is given, return the first match,
    otherwise return the first accepted item

    Return `None` if choices is given and no match is found,
    or nothing is specified.
    """
    for mime, params, q in _parse_accept_header(accept_header):
        if choices:
            if mime in choices:
                return mime
            else:
                continue
        else:
            return mime
    return None


def get_browser_protocol(request):
    """Get the _protocol_ seen by the browser

    Like tornado's _apply_xheaders,
    but in the case of multiple proxy hops,
    use the outermost value (what the browser likely sees)
    instead of the innermost value,
    which is the most trustworthy.

    We care about what the browser sees,
    not where the request actually came from,
    so trusting possible spoofs is the right thing to do.
    """
    headers = request.headers
    # first choice: Forwarded header
    forwarded_header = headers.get("Forwarded")
    if forwarded_header:
        first_forwarded = forwarded_header.split(",", 1)[0].strip()
        fields = {}
        for field in first_forwarded.split(";"):
            key, _, value = field.partition("=")
            fields[key.strip().lower()] = value.strip()
        if "proto" in fields and fields["proto"].lower() in {"http", "https"}:
            return fields["proto"].lower()
        else:
            app_log.warning(f"Forwarded header present without protocol: {forwarded_header}")

    # second choice: X-Scheme or X-Forwarded-Proto
    proto_header = headers.get("X-Scheme", headers.get("X-Forwarded-Proto", None))
    if proto_header:
        proto_header = proto_header.split(",")[0].strip().lower()
        if proto_header in {"http", "https"}:
            return proto_header

    # no forwarded headers
    return request.protocol


def convert_request_to_dict(arguments: Dict[str, List[bytes]]) -> Dict[str, Any]:
    """
    Converts the arguments obtained from a request to a dict.

    Args:
        handler: a tornado.web.RequestHandler object

    Returns:
        A decoded dict with keys/values extracted from the request's arguments
    """
    args = {}
    for k, values in arguments.items():
        args[k] = values[0].decode()
    return args



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import copy
import hashlib
import os
from typing import Optional

from kubernetes import config
from kubernetes.client.models import (
    V1Container,
    V1EnvVar,
    V1ObjectMeta,
    V1Pod,
    V1PodSecurityContext,
    V1PodSpec,
    V1ResourceRequirements,
    V1SecurityContext,
    V1Toleration,
    V1Volume,
    V1VolumeMount,
)


def get_current_namespace():
    ns_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
    if os.path.exists(ns_path):
        with open(ns_path) as f:
            return f.read().strip()
    try:
        _, active_context = config.list_kube_config_contexts()
        return active_context["context"]["namespace"]
    except KeyError:
        return "default"


def generate_hashed_slug(slug, limit=63, hash_length=6):
    """
    Generate a unique name that's within a certain length limit

    Most k8s objects have a 63 char name limit. We wanna be able to compress
    larger names down to that if required, while still maintaining some
    amount of legibility about what the objects really are.

    If the length of the slug is shorter than the limit - hash_length, we just
    return slug directly. If not, we truncate the slug to (limit - hash_length)
    characters, hash the slug and append hash_length characters from the hash
    to the end of the truncated slug. This ensures that these names are always
    unique no matter what.
    """
    if len(slug) < (limit - hash_length):
        return slug

    slug_hash = hashlib.sha256(slug.encode("utf-8")).hexdigest()

    return "{prefix}-{hash}".format(
        prefix=slug[: limit - hash_length - 1], hash=slug_hash[:hash_length]
    ).lower()


def update_k8s_model(target, changes, logger=None, target_name=None, changes_name=None):
    """
    Takes a model instance such as V1PodSpec() and updates it with another
    model, which is allowed to be a dict or another model instance of the same
    type. The logger is used to warn if any truthy value in the target is is
    overridden. The target_name parameter can for example be "pod.spec", and
    changes_name parameter could be "extra_pod_config". These parameters allows
    the logger to write out something more meaningful to the user whenever
    something is about to become overridden.
    """
    model_type = type(target)
    if not hasattr(target, "attribute_map"):
        raise AttributeError(
            "Attribute 'target' ({}) must be an object (such as 'V1PodSpec') "
            "with an attribute 'attribute_map'.".format(model_type.__name__)
        )
    if not isinstance(changes, model_type) and not isinstance(changes, dict):
        raise AttributeError(
            "Attribute 'changes' ({}) must be an object of the same type "
            "as 'target' ({}) or a 'dict'.".format(type(changes).__name__, model_type.__name__)
        )

    changes_dict = _get_k8s_model_dict(model_type, changes)
    for key, value in changes_dict.items():
        if key not in target.attribute_map:
            raise ValueError(
                "The attribute 'changes' ({}) contained '{}' not modeled by '{}'.".format(
                    type(changes).__name__, key, model_type.__name__
                )
            )

        # If changes are passed as a dict, they will only have a few keys/value
        # pairs representing the specific changes.
        # If the changes parameter is a model instance on the other hand,
        # the changes parameter will have a lot of default values as well.
        # These default values, which are also falsy,
        # should not use to override the target's values.
        if isinstance(changes, dict) or value:
            if getattr(target, key):
                if logger and changes_name:
                    warning = (
                        "'{}.{}' current value: '{}' "
                        "is overridden with '{}', "
                        "which is the value of '{}.{}'.".format(
                            target_name, key, getattr(target, key), value, changes_name, key
                        )
                    )
                    logger.warning(warning)
            setattr(target, key, value)

    return target


def get_k8s_model(model_type, model_dict):
    """
    Returns an instance of type specified model_type from an model instance or
    represantative dictionary.
    """
    model_dict = copy.deepcopy(model_dict)

    if isinstance(model_dict, model_type):
        return model_dict
    elif isinstance(model_dict, dict):
        # convert the dictionaries camelCase keys to snake_case keys
        model_dict = _map_dict_keys_to_model_attributes(model_type, model_dict)
        # use the dictionary keys to initialize a model of given type
        return model_type(**model_dict)
    else:
        raise AttributeError(
            "Expected object of type 'dict' (or '{}') but got '{}'.".format(
                model_type.__name__, type(model_dict).__name__
            )
        )


def _get_k8s_model_dict(model_type, model):
    """
    Returns a dictionary representation of a provided model type
    """
    model = copy.deepcopy(model)

    if isinstance(model, model_type):
        return model.to_dict()
    elif isinstance(model, dict):
        return _map_dict_keys_to_model_attributes(model_type, model)
    else:
        raise AttributeError(
            "Expected object of type '{}' (or 'dict') but got '{}'.".format(
                model_type.__name__, type(model).__name__
            )
        )


def _map_dict_keys_to_model_attributes(model_type, model_dict):
    """
    Maps a dict's keys to the provided models attributes
    using its attribute_map attribute.
    This is (always?) the same as converting camelCase to snake_case.
    Note that the function will not influence nested object's keys.
    """

    new_dict = {}
    for key, value in model_dict.items():
        new_dict[_get_k8s_model_attribute(model_type, key)] = value

    return new_dict


def _get_k8s_model_attribute(model_type, field_name):
    """
    Takes a model type and a Kubernetes API resource field name (such as
    "serviceAccount") and returns a related attribute name (such as
    "service_account") to be used with  kubernetes.client.models objects. It is
    impossible to prove a negative but it seems like it is always a question of
    making camelCase to snake_case but by using the provided 'attribute_map' we
    also ensure that the fields actually exist.

    Example of V1PodSpec's attribute_map:
    {
        'active_deadline_seconds': 'activeDeadlineSeconds',
        'affinity': 'affinity',
        'automount_service_account_token': 'automountServiceAccountToken',
        'containers': 'containers',
        'dns_policy': 'dnsPolicy',
        'host_aliases': 'hostAliases',
        'host_ipc': 'hostIPC',
        'host_network': 'hostNetwork',
        'host_pid': 'hostPID',
        'hostname': 'hostname',
        'image_pull_secrets': 'imagePullSecrets',
        'init_containers': 'initContainers',
        'node_name': 'nodeName',
        'node_selector': 'nodeSelector',
        'priority': 'priority',
        'priority_class_name': 'priorityClassName',
        'restart_policy': 'restartPolicy',
        'scheduler_name': 'schedulerName',
        'security_context': 'securityContext',
        'service_account': 'serviceAccount',
        'service_account_name': 'serviceAccountName',
        'subdomain': 'subdomain',
        'termination_grace_period_seconds': 'terminationGracePeriodSeconds',
        'tolerations': 'tolerations',
        'volumes': 'volumes'
    }
    """
    # if we get "service_account", return
    if field_name in model_type.attribute_map:
        return field_name

    # if we get "serviceAccount", then return "service_account"
    for key, value in model_type.attribute_map.items():
        if value == field_name:
            return key
    else:
        raise ValueError(
            "'{}' did not have an attribute matching '{}'".format(model_type.__name__, field_name)
        )


def make_pod(
    name: str,
    cmd: list[str],
    env: list[V1EnvVar],
    image: str,
    image_pull_policy: str,
    image_pull_secrets: Optional[list] = None,
    working_dir: Optional[str] = None,
    volumes: Optional[list] = None,
    volume_mounts: Optional[list] = None,
    labels: Optional[dict] = None,
    annotations: Optional[dict] = None,
    node_selector: Optional[dict] = None,
    tolerations: Optional[list] = None,
    run_as_user: Optional[int] = None,
) -> V1Pod:
    """
    Creates a Kubernetes Pod specification (V1Pod) with the given parameters.

    Args:
        name (str): The name of the pod.
        cmd (list[str]): The command to run in the pod.
        image (str): The Docker image to use for the pod's container.
        image_pull_policy (str): The image pull policy.
        image_pull_secrets (Optional[list], default None): List of secrets to pull images from private registries.
        working_dir (Optional[str], default None): The working directory for the container.
        volumes (Optional[list], default None): List of volumes to attach to the pod.
        volume_mounts (Optional[list], default None): List of volume mounts for the container.
        labels (Optional[dict], default None): Labels to associate with the pod.
        annotations (Optional[dict], default None): Annotations to associate with the pod.
        node_selector (Optional[dict], default None): Node selectors to determine where the pod should be scheduled.
        tolerations (Optional[list], default None): Tolerations to apply for the pod.
        run_as_user (Optional[int], default None): The user ID under which the container should run.

    Returns:
        V1Pod: The Kubernetes Pod specification object.
    """
    pod = V1Pod()
    pod.kind = "Pod"
    pod.api_version = "v1"

    pod.metadata = V1ObjectMeta(
        name=name, labels=(labels or {}).copy(), annotations=(annotations or {}).copy()
    )

    pod.spec = V1PodSpec(
        containers=[],
        security_context=V1PodSecurityContext(),
        image_pull_secrets=image_pull_secrets,
        restart_policy="Never",
        node_selector=node_selector,
    )
    # TODO maybe get userid of jupyterhub user
    autograde_container = V1Container(
        name="autograde",
        image=image,
        working_dir=working_dir,
        args=cmd,
        env=env,
        image_pull_policy=image_pull_policy,
        resources=V1ResourceRequirements(),
        security_context=V1SecurityContext(run_as_user=run_as_user),
        volume_mounts=[get_k8s_model(V1VolumeMount, obj) for obj in (volume_mounts or [])],
    )

    pod.spec.containers.append(autograde_container)
    if tolerations:
        pod.spec.tolerations = [get_k8s_model(V1Toleration, obj) for obj in tolerations]
    if volumes:
        pod.spec.volumes = [get_k8s_model(V1Volume, obj) for obj in volumes]

    return pod



import os
from typing import Any

from traitlets import List

from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.convert import utils
from grader_service.convert.converters.base import BaseConverter
from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.gradebook.gradebook import Gradebook, MissingEntry
from grader_service.convert.preprocessors import (
    CheckCellMetadata,
    ClearAlwaysHiddenTests,
    ClearOutput,
    DeduplicateIds,
    Execute,
    LimitOutput,
    OverwriteCells,
    OverwriteKernelspec,
    SaveAutoGrades,
)


class Autograde(BaseConverter):
    _sanitizing = True

    sanitize_preprocessors = List(
        [ClearOutput, DeduplicateIds, OverwriteKernelspec, OverwriteCells, CheckCellMetadata]
    ).tag(config=True)
    autograde_preprocessors = List(
        [Execute, LimitOutput, SaveAutoGrades, CheckCellMetadata, ClearAlwaysHiddenTests]
    ).tag(config=True)

    preprocessors = List([])

    def _init_preprocessors(self) -> None:
        self.exporter._preprocessors = []
        if self._sanitizing:
            preprocessors = self.sanitize_preprocessors
        else:
            preprocessors = self.autograde_preprocessors

        for pp in preprocessors:
            self.exporter.register_preprocessor(pp)

    def convert_single_notebook(self, notebook_filename: str) -> None:
        # ignore notebooks that aren't in the gradebook
        resources = self.init_single_notebook_resources(notebook_filename)
        with Gradebook(resources["output_json_path"]) as gb:
            try:
                gb.find_notebook(resources["unique_key"])
            except MissingEntry:
                self.log.warning("Skipping unknown notebook: %s", notebook_filename)
                return

        self.log.info("Sanitizing %s", notebook_filename)
        self._sanitizing = True
        self._init_preprocessors()
        super().convert_single_notebook(notebook_filename)

        notebook_filename = os.path.join(
            self.writer.build_directory, os.path.basename(notebook_filename)
        )
        self.log.info("Autograding %s", notebook_filename)
        self._sanitizing = False
        self._init_preprocessors()
        try:
            with utils.setenv(NBGRADER_EXECUTION="autograde"):
                super().convert_single_notebook(notebook_filename)
        finally:
            self._sanitizing = True

    def convert_notebooks(self) -> None:
        # check for missing notebooks and give them a score of zero if they do not exist
        json_path = os.path.join(self._output_directory, "gradebook.json")
        with Gradebook(json_path) as gb:
            glob_notebooks = {
                self.init_single_notebook_resources(n)["unique_key"]: n for n in self.notebooks
            }
            for notebook in gb.model.notebook_id_set.difference(set(glob_notebooks.keys())):
                self.log.warning("No submitted file: %s", notebook)
                nb = gb.find_notebook(notebook)
                for grade in nb.grades:
                    grade.auto_score = 0
                    grade.needs_manual_grade = False
                    gb.add_grade(grade.id, notebook, grade)

        super().convert_notebooks()

    def __init__(
        self,
        input_dir: str,
        output_dir: str,
        file_pattern: str,
        assignment_settings: AssignmentSettings,
        **kwargs: Any,
    ) -> None:
        super().__init__(input_dir, output_dir, file_pattern, assignment_settings, **kwargs)
        self.force = True  # always overwrite generated assignments


class AutogradeApp(ConverterApp):
    version = ConverterApp.__version__

    def start(self):
        Autograde(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            assignment_settings=utils.get_assignment_settings_from_env(),
            config=self.config,
        ).start()



"""add user display name

Revision ID: 28500016a3c3
Revises: fc5d2febe781
Create Date: 2025-04-08 12:10:09.318559

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "28500016a3c3"
down_revision = "fc5d2febe781"
branch_labels = None
depends_on = None


def upgrade():
    dialect = op.get_bind().dialect.name
    if dialect == "sqlite":
        # sqlite has to recreate the tables on `batch_alter_table`, but dropping a table
        # would cause integrity errors, so we disable the foreign key constraint temporarily
        op.execute(sa.text("PRAGMA foreign_keys=OFF"))

    # Step 1: Add column as nullable
    op.add_column("user", sa.Column("display_name", sa.String(), nullable=True))

    # Step 2: Copy data from 'name' to 'display_name'
    user_table = sa.table(
        "user", sa.column("name", sa.String), sa.column("display_name", sa.String)
    )
    op.execute(user_table.update().values(display_name=user_table.c.name))

    # Step 3: Make the column non-nullable
    # SQLite needs batch mode; PostgreSQL can handle direct alter
    with op.batch_alter_table("user") as batch_op:
        batch_op.alter_column("display_name", nullable=False)


def downgrade():
    op.drop_column("user", "display_name")



from traitlets import Unicode, default

from grader_service.auth.auth import Authenticator
from grader_service.auth.login import LogoutHandler


class DummyAuthenticator(Authenticator):
    """Dummy Authenticator for testing

    By default, any username + password is allowed
    If a non-empty password is set, any username will be allowed
    if it logs in with that password.

    """

    logout_handler = LogoutHandler

    @default("allow_all")
    def _allow_all_default(self):
        if self.allowed_users:
            return False
        else:
            # allow all by default
            return True

    password = Unicode(
        config=True,
        help="""
        Set a global password for all users wanting to log in.

        This allows users with any username to log in with the same static password.
        """,
    )

    def check_allow_config(self):
        super().check_allow_config()
        self.log.warning(
            f"Using testing authenticator {self.__class__.__name__}! This is not meant for production!"
        )

    async def authenticate(self, handler, data):
        """Checks against a global password if it's been set. If not, allow any user/pass combo"""
        if self.password:
            if data["password"] == self.password:
                return data["username"]
            return None
        return data["username"]

    def get_handlers(self, base_url_path: str):
        base_handlers = super().get_handlers(base_url_path)
        dummy_handlers = [(self.logout_url(base_url_path), self.logout_handler)]
        return base_handlers + dummy_handlers



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import enum
from http import HTTPStatus
from typing import Union

from tornado.web import HTTPError


def parse_ids(*args) -> Union[int, tuple[int, ...]]:
    """
    Validates that provided args are all integers.

    :param args: certain amount of id (int) values
    :return: tuple of ids as ints, or a single id if only one was provided
    :raises HTTPError: if args contain invalid ids which cannot be cast to int
    """
    try:
        ids = tuple(int(id) for id in args)
    except ValueError:
        raise HTTPError(HTTPStatus.BAD_REQUEST, reason="All IDs have to be numerical")
    if len(ids) == 1:
        return ids[0]
    return ids


class GitRepoType(enum.StrEnum):
    """Allowed repository types.

    SOURCE: The source files created by the instructor.
    RELEASE: The "student" version of the source files.
    USER: A user's copy of the release files, which can be submitted.
    AUTOGRADE: Autograded submission files.
    EDIT: Copy of the submission files, created by the instructor for manual editing.
    FEEDBACK: Final feedback for a submission.
    """

    SOURCE = "source"
    RELEASE = "release"
    USER = "user"
    AUTOGRADE = "autograde"
    EDIT = "edit"
    FEEDBACK = "feedback"



"""Add feedback status

Revision ID: 9cfeb0faa0c0
Revises: 4620c13728e7
Create Date: 2024-01-10 15:16:36.273796

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9cfeb0faa0c0"
down_revision = "4620c13728e7"
branch_labels = None
depends_on = None


def upgrade():
    sa_enum = sa.Enum(
        "not_generated",
        "generated",
        "generating",
        "generation_failed",
        "feedback_outdated",
        name="feedback_status",
    )
    # This will create the enum if it's not already present
    sa_enum.create(op.get_bind(), checkfirst=True)
    op.add_column(
        "submission",
        sa.Column("feedback_status", sa_enum, server_default="not_generated", nullable=False),
    )

    op.execute(
        "UPDATE submission SET feedback_status='not_generated' WHERE feedback_available=false"
    )
    op.execute("UPDATE submission SET feedback_status='generated' WHERE feedback_available=true")
    op.drop_column("submission", "feedback_available")


def downgrade():
    op.add_column(
        "submission",
        sa.Column("feedback_available", sa.Boolean, server_default="false", nullable=False),
    )
    op.execute(
        "UPDATE submission SET feedback_available=false WHERE feedback_status='not_generated' OR feedback_status='generation_failed' OR feedback_status='generating'"
    )
    op.execute(
        "UPDATE submission SET feedback_available=true WHERE feedback_status='generated' OR feedback_status='feedback_outdated'"
    )
    op.drop_column("submission", "feedback_status")

    sa_enum = sa.Enum(name="feedback_status")
    sa_enum.drop(op.get_bind(), checkfirst=True)



from dataclasses import asdict, dataclass
from enum import Enum
from typing import Dict, List, Optional, Set, Type, Union


@dataclass
class BaseModel:
    def __post_init__(self):
        self._type = self.__class__.__name__

    @classmethod
    def empty_dict(cls: Type["BaseModel"]) -> dict:
        fields = cls.__dataclass_fields__.keys()
        return {f: None for f in fields}

    def to_dict(self) -> dict:
        d = asdict(self)
        d["_type"] = self.__class__.__name__
        return d

    @classmethod
    def from_dict(cls: Type["BaseModel"], d: dict) -> Type["BaseModel"]:
        if d is None or len(d) == 0:
            return None
        d_no_type = {k: v for k, v in d.items() if k != "_type"}
        return cls(**d_no_type)


@dataclass
class IDMixin:
    id: str


@dataclass
class NameMixin:
    name: str


@dataclass
class NotebookRelashionship:
    notebook_id: str


@dataclass
class CellRelashionship:
    cell_id: str


@dataclass
class Grade(BaseModel, IDMixin, NotebookRelashionship, CellRelashionship):
    """Representation of a grade assigned to the submitted version of a grade cell."""

    auto_score: float
    manual_score: float
    extra_credit: float
    needs_manual_grade: bool
    #: The maximum possible score that can be assigned, inherited from :class:`~GradeCell`
    max_score_gradecell: float
    max_score_taskcell: float

    #: Whether the autograded score is a result of failed autograder tests. This
    #: is True if the autograder score is zero and the cell type is "code", and
    #: otherwise False.
    failed_tests: bool

    @property
    def score(self) -> float:
        """
        The overall score, computed automatically from the
        :attr:`~models.Grade.auto_score` and :attr:`~models.Grade.manual_score`
        values. If neither are set, the score is zero. If both are set, then the
        manual score takes precedence. If only one is set, then that value is used
        for the score
        """
        if self.manual_score is None and self.auto_score is None:
            if self.extra_credit is None:
                return 0.0
            else:
                return self.extra_credit
        elif self.manual_score is None:
            if self.extra_credit is None:
                return self.auto_score
            else:
                return self.auto_score + self.extra_credit
        else:
            if self.extra_credit is None:
                return self.manual_score
            else:
                return self.manual_score + self.extra_credit

    @property
    def max_score(self):
        if self.max_score_taskcell:
            return self.max_score_taskcell
        else:
            return self.max_score_gradecell

    @classmethod
    def from_dict(cls: Type["Grade"], d: dict) -> Type["Grade"]:
        if d is None or len(d) == 0:
            return None
        d = {k: v for k, v in d.items() if k not in ["_type", "max_score"]}
        return cls(**d)

    def to_dict(self) -> dict:
        d = super().to_dict()
        d["max_score"] = self.max_score
        return d


@dataclass
class Comment(BaseModel, IDMixin, NotebookRelashionship, CellRelashionship):
    auto_comment: str
    manual_comment: str

    @property
    def comment(self) -> Optional[str]:
        if self.manual_comment is None:
            return self.auto_comment
        else:
            return self.manual_comment


class CellType(Enum):
    "code"

    "markdown"


@dataclass
class BaseCell(BaseModel, IDMixin, NameMixin, NotebookRelashionship):
    grade_id: str  # we can have only one grade
    comment_id: str  # we can have only one comment since we only process a single notebook

    @classmethod
    def from_dict(cls: Type["BaseCell"], d: dict) -> Type["BaseCell"]:
        if d["_type"] == "GradeCell":
            return GradeCell.from_dict(d)
        elif d["_type"] == "SolutionCell":
            return SolutionCell.from_dict(d)
        elif d["_type"] == "TaskCell":
            return TaskCell.from_dict(d)
        elif d["_type"] == "SourceCell":
            return SourceCell.from_dict(d)


@dataclass
class GradedMixin:
    max_score: float
    cell_type: CellType


@dataclass
class GradeCell(BaseCell, GradedMixin):
    @classmethod
    def from_dict(cls: Type["BaseCell"], d: dict) -> "GradeCell":
        return GradeCell(
            max_score=d["max_score"],
            cell_type=d["cell_type"],
            id=d["id"],
            notebook_id=d["notebook_id"],
            name=d["name"],
            grade_id=d["grade_id"],
            comment_id=d["comment_id"],
        )


@dataclass
class SolutionCell(BaseCell):
    @classmethod
    def from_dict(cls: Type["BaseCell"], d: dict) -> "SolutionCell":
        return SolutionCell(
            id=d["id"],
            notebook_id=d["notebook_id"],
            name=d["name"],
            grade_id=d["grade_id"],
            comment_id=d["comment_id"],
        )


@dataclass
class TaskCell(BaseCell, GradedMixin):
    @classmethod
    def from_dict(cls: Type["BaseCell"], d: dict) -> "TaskCell":
        return TaskCell(
            max_score=d["max_score"],
            cell_type="code",  # cell_type from GradedMixin should always be 'code'
            id=d["id"],
            notebook_id=d["notebook_id"],
            name=d["name"],
            grade_id=d["grade_id"],
            comment_id=d["comment_id"],
        )


@dataclass
class SourceCell(BaseModel, IDMixin, NameMixin, NotebookRelashionship):
    cell_type: CellType

    #: Whether the cell is locked (e.g. the source saved in the database should
    #: be used to overwrite the source of students' cells)
    locked: bool

    #: The source code or text of the cell
    source: str

    #: A checksum of the cell contents. This should usually be computed
    #: using :func:`utils.compute_checksum`
    checksum: str

    @classmethod
    def from_dict(cls: Type["BaseModel"], d: dict) -> Type["BaseModel"]:
        return super().from_dict(d)


@dataclass
class Notebook(BaseModel, IDMixin, NameMixin):
    kernelspec: str
    grade_cells_dict: Dict[str, Type[GradeCell]]
    solution_cells_dict: Dict[str, Type[SolutionCell]]
    task_cells_dict: Dict[str, Type[TaskCell]]
    source_cells_dict: Dict[str, Type[SourceCell]]
    grades_dict: Dict[str, Type[Grade]]
    comments_dict: Dict[str, Type[Comment]]
    #: Whether this assignment has been flagged by a human grader
    flagged: bool

    def __post_init__(self):
        super().__post_init__()
        if self.flagged is None:
            self.flagged = False

    @property
    def grade_cells(self) -> List[GradeCell]:
        return [x for x in self.grade_cells_dict.values() if isinstance(x, GradeCell)]

    @property
    def solution_cells(self) -> List[SolutionCell]:
        return [x for x in self.solution_cells_dict.values() if isinstance(x, SolutionCell)]

    @property
    def task_cells(self) -> List[TaskCell]:
        return [x for x in self.task_cells_dict.values() if isinstance(x, TaskCell)]

    @property
    def source_cells(self) -> List[SourceCell]:
        return [x for x in self.source_cells_dict.values()]

    @property
    def graded_cells(self) -> List[Union[GradeCell, TaskCell]]:
        return self.grade_cells + self.task_cells

    @property
    def max_score(self) -> float:
        return sum([g.max_score for g in self.graded_cells])

    @property
    def score(self) -> float:
        return sum([gr.score for gr in self.grades_dict.values()])

    @property
    def code_score(self) -> float:
        return sum(
            [
                self.grades_dict[g.grade_id].score
                for g in self.graded_cells
                if g.cell_type == "code" and g is not None and g.grade_id is not None
            ]
        )

    @property
    def max_code_score(self) -> float:
        return sum([g.max_score for g in self.graded_cells if g.cell_type == "code"])

    @property
    def written_score(self) -> float:
        return sum(
            [
                self.grades_dict[g.grade_id].score
                for g in self.graded_cells
                if g.cell_type == "markdown"
            ]
        )

    @property
    def max_written_score(self) -> float:
        return sum([g.max_score for g in self.graded_cells if g.cell_type == "markdown"])

    @property
    def failed_tests(self) -> bool:
        return any([g.failed_tests for g in self.grades])

    @property
    def grades(self) -> List[Type[Grade]]:
        return list(self.grades_dict.values())

    @property
    def comments(self) -> List[Type[Comment]]:
        return list(self.comments_dict.values())

    @classmethod
    def from_dict(cls: Type["Notebook"], d: dict) -> "Notebook":
        gc = {id: GradeCell.from_dict(v) for id, v in d["grade_cells_dict"].items()}
        sc = {i: SolutionCell.from_dict(v) for i, v in d["solution_cells_dict"].items()}
        tc = {id: TaskCell.from_dict(v) for id, v in d["task_cells_dict"].items()}
        sr = {id: SourceCell.from_dict(v) for id, v in d["source_cells_dict"].items()}
        gr = {id: Grade.from_dict(v) for id, v in d["grades_dict"].items()}
        co = {id: Comment.from_dict(v) for id, v in d["comments_dict"].items()}
        return Notebook(
            kernelspec=d["kernelspec"],
            grade_cells_dict=gc,
            solution_cells_dict=sc,
            task_cells_dict=tc,
            id=d["id"],
            name=d["name"],
            flagged=d["flagged"],
            source_cells_dict=sr,
            grades_dict=gr,
            comments_dict=co,
        )

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "flagged": self.flagged,
            "kernelspec": self.kernelspec,
            "grade_cells_dict": {k: v.to_dict() for k, v in self.grade_cells_dict.items()},
            "solution_cells_dict": {k: v.to_dict() for k, v in self.solution_cells_dict.items()},
            "task_cells_dict": {k: v.to_dict() for k, v in self.task_cells_dict.items()},
            "source_cells_dict": {k: v.to_dict() for k, v in self.source_cells_dict.items()},
            "grades_dict": {k: v.to_dict() for k, v in self.grades_dict.items()},
            "comments_dict": {k: v.to_dict() for k, v in self.comments_dict.items()},
            "_type": self._type,
        }


@dataclass
class GradeBookModel(BaseModel):
    notebooks: Dict[str, Notebook]
    extra_files: List[str]

    def __str__(self):
        return f"GradeBookModel: notebooks={list(self.notebooks.keys())}, extra_files={self.extra_files}"

    @property
    def score(self) -> float:
        score = 0
        for nb in self.notebooks.values():
            score += nb.score
        return score

    @property
    def max_score(self) -> float:
        max_score = 0
        for nb in self.notebooks.values():
            max_score += nb.max_score
        return max_score

    @property
    def notebook_id_set(self) -> Set[str]:
        return {x for x in self.notebooks.keys()}

    @classmethod
    def from_dict(cls: Type["GradeBookModel"], d: dict) -> "GradeBookModel":
        ns = {id: Notebook.from_dict(v) for id, v in d["notebooks"].items()}
        extra_files = d.get("extra_files", [])
        return GradeBookModel(notebooks=ns, extra_files=extra_files)

    def to_dict(self) -> dict:
        return {
            "notebooks": {k: v.to_dict() for k, v in self.notebooks.items()},
            "extra_files": self.extra_files,
            "_type": self._type,
            "schema_version": "1",
        }



import os

## generic
c.JupyterHub.admin_access = True
c.Spawner.default_url = "/lab"
c.Spawner.cmd = ["jupyter-labhub"]

## authenticator
from oauthenticator.generic import GenericOAuthenticator

c.JupyterHub.authenticator_class = GenericOAuthenticator
c.GenericOAuthenticator.oauth_callback_url = "http://localhost:8080/hub/oauth_callback"

c.GenericOAuthenticator.client_id = "my_id"
c.GenericOAuthenticator.client_secret = "my_secret"
c.GenericOAuthenticator.authorize_url = "http://localhost:4010/services/grader/api/oauth2/authorize"
c.GenericOAuthenticator.token_url = "http://localhost:4010/services/grader/api/oauth2/token"
c.GenericOAuthenticator.logout_redirect_url = "http://localhost:4010/services/grader/logout"

c.GenericOAuthenticator.userdata_url = "http://localhost:4010/services/grader/api/user"
c.GenericOAuthenticator.username_claim = "name"
c.Authenticator.enable_auth_state = True

##############################
# PASS GRADER TOKEN TO SPAWNER


def auth_state_hook(spawner, auth_state):
    token = auth_state["access_token"]

    # The environment variable GRADER_API_TOKEN is used by the lab-extension
    # to identify the user in API calls to the Grader Service.
    spawner.environment.update({"GRADER_API_TOKEN": token})


# We have access to the authentication data, which we can use to set
# `userdata` in the spawner of the user.
c.Spawner.auth_state_hook = auth_state_hook

##############################

c.Authenticator.allowed_users = {"admin", "instructor", "tutor", "student"}
c.Authenticator.admin_users = {"admin"}

## spawner
c.JupyterHub.spawner_class = "jupyterhub.spawner.SimpleLocalProcessSpawner"
c.SimpleLocalProcessSpawner.home_dir_template = os.path.join(os.getcwd(), "home_dir", "{username}")

## simple setup
c.JupyterHub.ip = "127.0.0.1"
c.JupyterHub.port = 8080

c.JupyterHub.services.append(
    {
        "name": "grader",
        "url": "http://127.0.0.1:4010",
        "api_token": "7572f93a2e7640999427d9289c8318c0",
    }
)

c.JupyterHub.log_level = "INFO"



import base64
import hashlib
import json
import re
import uuid
from typing import Any, Dict, Optional, cast
from urllib.parse import quote, unquote

from oauthlib.common import generate_token
from tornado.httputil import url_concat
from tornado.log import app_log
from tornado.web import HTTPError, MissingArgumentError, RequestHandler

from grader_service.handlers.base_handler import BaseHandler
from grader_service.utils import convert_request_to_dict, url_path_join

from .error import InvalidAudienceError, LoginError, ValidationError
from .validator import LTI13LaunchValidator

STATE_COOKIE_NAME = "grader-lti13authenticator-state"
NONCE_STATE_COOKIE_NAME = "grader-lti13authenticator-nonce-state"


def make_nonce_state() -> str:
    """
    Create state for nonce calculation.
    """
    return generate_token(length=64)


def get_nonce(nonce_state: str) -> str:
    """
    Create a nonce by hashing state.

    SHA 256 is used to create the hash. The nonce is its hexdigest.
    """
    hash = hashlib.sha256(nonce_state.encode())
    nonce = hash.hexdigest()
    return nonce


def _serialize_state(state):
    """Serialize OAuth state to a base64 string after passing through JSON"""
    json_state = json.dumps(state)
    return base64.urlsafe_b64encode(json_state.encode("utf8")).decode("ascii")


def _deserialize_state(b64_state):
    """Deserialize OAuth state as serialized in _serialize_state"""
    if isinstance(b64_state, str):
        b64_state = b64_state.encode("ascii")
    try:
        json_state = base64.urlsafe_b64decode(b64_state).decode("utf8")
    except ValueError:
        app_log.error(f"Failed to b64-decode state: {b64_state}")
        return {}
    try:
        return json.loads(json_state)
    except ValueError:
        app_log.error(f"Failed to json-decode state: {json_state}")
        return {}


class LTI13ConfigHandler(BaseHandler):
    """
    Handles JSON configuration file for LTI 1.3.
    """

    async def get(self) -> None:
        """
        Gets the JSON config which is used by LTI platforms
        to install the external tool.

        - The extensions key contains settings for specific vendors, such as canvas,
        moodle, edx, among others.
        - The tool uses public settings by default. Users that wish to install the tool with
        private settings should either copy/paste the json or toggle the application to private
        after it is installed with the platform.
        - Usernames are obtained by first attempting to get and normalize values sent when
        tools are installed with public settings. If private, the username is set using the
        anonumized user data when requests are sent with private installation settings.
        """
        self.set_header("Content-Type", "application/json")

        # get the origin protocol
        protocol = self.authenticator.get_uri_scheme(self.request)
        self.log.debug(f"Origin protocol is: {protocol}")
        # build the full target link url value required for the jwks endpoint
        target_link_url = f"{protocol}://{self.request.host}"
        self.log.debug(f"Target link url is: {target_link_url}")
        keys = {
            "title": self.authenticator.tool_name,
            "scopes": [
                "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem",
                "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem.readonly",
                "https://purl.imsglobal.org/spec/lti-ags/scope/result.readonly",
                "https://purl.imsglobal.org/spec/lti-ags/scope/score",
                "https://purl.imsglobal.org/spec/lti-nrps/scope/contextmembership.readonly",
                "https://canvas.instructure.com/lti/public_jwk/scope/update",
                "https://canvas.instructure.com/lti/data_services/scope/create",
                "https://canvas.instructure.com/lti/data_services/scope/show",
                "https://canvas.instructure.com/lti/data_services/scope/update",
                "https://canvas.instructure.com/lti/data_services/scope/list",
                "https://canvas.instructure.com/lti/data_services/scope/destroy",
                "https://canvas.instructure.com/lti/data_services/scope/list_event_types",
                "https://canvas.instructure.com/lti/feature_flags/scope/show",
                "https://canvas.instructure.com/lti/account_lookup/scope/show",
            ],
            "extensions": [
                {
                    "platform": "canvas.instructure.com",
                    "settings": {
                        "platform": "canvas.instructure.com",
                        "placements": [
                            {
                                "placement": "course_navigation",
                                "message_type": "LtiResourceLinkRequest",
                                "windowTarget": "_blank",
                                "target_link_uri": target_link_url,
                                "custom_fields": {
                                    "email": "$Person.email.primary",
                                    "lms_user_id": "$User.id",
                                },
                            },
                            {
                                "placement": "assignment_selection",
                                "message_type": "LtiResourceLinkRequest",
                                "target_link_uri": target_link_url,
                            },
                        ],
                    },
                    "privacy_level": "public",
                }
            ],
            "description": self.authenticator.tool_description,
            "custom_fields": {"email": "$Person.email.primary", "lms_user_id": "$User.id"},
            "target_link_uri": target_link_url,
            "oidc_initiation_url": self.authenticator.login_url(
                url_path_join(target_link_url, self.hub.server.base_url)
            ),
        }
        self.write(json.dumps(keys))


class LTI13LoginInitHandler(BaseHandler):
    """
    Handles JupyterHub authentication requests according to the
    LTI 1.3 standard.
    """

    _state = None

    def check_xsrf_cookie(self):
        """
        Do not attempt to check for xsrf parameter in POST requests. LTI requests are
        meant to be cross-site, so it must not be verified.
        """
        return

    def authorize_redirect(
        self,
        redirect_uri: str,
        login_hint: str,
        nonce: str,
        client_id: str,
        state: str,
        lti_message_hint: Optional[str] = None,
    ) -> None:
        """
        Overrides the OAuth2Mixin.authorize_redirect method to to initiate the LTI 1.3 / OIDC
        login flow with the required and optional arguments.

        User Agent (browser) is redirected to the platform's authorization url for further
        processing.

        References:
        https://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
        http://www.imsglobal.org/spec/lti/v1p3/#additional-login-parameters-0

        Args:
          redirect_uri: redirect url specified during tool installation (callback url) to
            which the user will be redirected from the platform after attempting authorization.
          login_hint: opaque value used by the platform for user identity
          nonce: unique value sent to allow recipients to protect themselves against replay attacks
          client_id: used to identify the tool's installation with a platform
          state: opaque value for the platform to maintain state between the request and
            callback and provide Cross-Site Request Forgery (CSRF) mitigation.
          lti_message_hint: similarly to the login_hint parameter, lti_message_hint value is opaque to the tool.
            If present in the login initiation request, the tool MUST include it back in
            the authentication request unaltered.
        """
        handler = cast(RequestHandler, self)
        # Required parameter with values specified by LTI 1.3
        # https://www.imsglobal.org/spec/security/v1p0/#step-2-authentication-request
        args = {
            "response_type": "id_token",
            "scope": "openid",
            "response_mode": "form_post",
            "prompt": "none",
        }
        # Dynamically computed required parameter values
        args["client_id"] = client_id
        args["redirect_uri"] = redirect_uri
        args["login_hint"] = login_hint
        args["nonce"] = nonce
        args["state"] = state

        if lti_message_hint is not None:
            args["lti_message_hint"] = lti_message_hint

        url = self.authenticator.authorize_url
        handler.redirect(url_concat(url, args))

    def get_state(self):
        next_url = original_next_url = self.get_argument("next", None)
        if not next_url:
            # try with the target_link_uri arg
            target_link = self.get_argument("target_link_uri", "")
            if "next" in target_link:
                self.log.debug(f"Trying to get the next-url from target_link_uri: {target_link}")
                next_search = re.search("next=(.*)", target_link, re.IGNORECASE)
                if next_search:
                    next_url = next_search.group(1)
                    # decode the some characters obtained with the link builder
                    next_url = unquote(next_url)
            elif not target_link.endswith("/hub"):
                next_url = target_link
        if next_url:
            # Note: allow hostname-having urls
            # avoid browsers treating \ as /
            next_url = next_url.replace("\\", quote("\\"))
            if next_url != original_next_url:
                self.log.warning("Ignoring next_url %r, using %r", original_next_url, next_url)
        if self._state is None:
            self._state = _serialize_state({"state_id": uuid.uuid4().hex, "next_url": next_url})
        return self._state

    async def post(self):
        """
        Validates required login arguments sent from platform and then uses the authorize_redirect() method
        to redirect users to the authorization url.
        """
        validator = LTI13LaunchValidator()
        args = convert_request_to_dict(self.request.arguments)
        self.log.debug(f"Initial login request args are {args}")

        # Raises HTTP 400 if login request arguments are not valid
        try:
            validator.validate_login_request(args)
        except ValidationError as e:
            self.log.error(str(e))
            html = await self.render_template("auth/lti_no_auth.html.j2")
            await self.finish(html)
            return

        login_hint = args["login_hint"]
        self.log.debug(f"login_hint is {login_hint}")

        lti_message_hint = self._get_optional_arg(args, "lti_message_hint")
        client_id = self._get_optional_arg(args, "client_id")

        # lti_deployment_id is not used anywhere. It may be used in the future to influence the
        # login flow depending on the deployment settings. A configurable hook, similar to `Authenticator`'s `post_auth_hook`
        # would be a good way to implement this.
        # lti_deployment_id = self._get_optional_arg(args, "lti_deployment_id")

        redirect_uri = self.get_redirect_uri()
        self.log.debug(f"redirect_uri is: {redirect_uri}")

        # to prevent CSRF
        state = self.generate_state()

        # to prevent replay attacks
        nonce = self.generate_nonce()
        self.log.debug(f"nonce value: {nonce}")

        self.authorize_redirect(
            client_id=client_id,
            login_hint=login_hint,
            lti_message_hint=lti_message_hint,
            nonce=nonce,
            redirect_uri=redirect_uri,
            state=state,
        )

    # GET requests are also allowed by the OpenID Connect launch flow:
    # https://www.imsglobal.org/spec/security/v1p0/#fig_oidcflow
    #
    get = post

    def generate_state(self):
        """Produce a state including the url of the original request."""
        state = self.get_state()
        self.set_state_cookie(state)
        return state

    def generate_nonce(self):
        """Produce a nonce.

        The nonce state will be stored as a session cookie to later validate the nonce
        field of the id_token.
        """
        nonce_state = make_nonce_state()
        self.set_nonce_state_cookie(nonce_state)
        nonce = get_nonce(nonce_state)
        return nonce

    def get_redirect_uri(self) -> str:
        """Create uri to redirect user agent to after successful authorization by the LMS platform."""
        return "{proto}://{host}{path}".format(
            proto=self.authenticator.get_uri_scheme(self.request),
            host=self.request.host,
            path=self.authenticator.callback_url(self.application.base_url),
        )

    def _get_optional_arg(self, args: Dict[str, str], arg: str) -> Optional[str]:
        """
        Return value of optional argument or None if not present.
        """
        value = args.get(arg)
        if value:
            self.log.debug(f"{arg} is {value}")
        else:
            self.log.debug(f"{arg} not present in login initiation request")
        return value

    def set_nonce_state_cookie(self, nonce_state):
        self._set_oauth_cookie(NONCE_STATE_COOKIE_NAME, nonce_state)

    def set_state_cookie(self, state):
        self._set_oauth_cookie(STATE_COOKIE_NAME, state)

    def _set_oauth_cookie(self, key: str, value):
        self._set_cookie(key, value, expires_days=1, httponly=True, encrypted=True)


class LTI13CallbackHandler(BaseHandler):
    """
    Handles JupyterHub authentication requests responses according to the
    LTI 1.3 standard.

    References:
    https://www.imsglobal.org/spec/security/v1p0/#step-3-authentication-response
    https://www.imsglobal.org/spec/security/v1p0/#step-4-resource-is-displayed
    """

    _state_cookie = None
    _nonce_state_cookie = None

    def check_xsrf_cookie(self):
        """
        Do not attempt to check for xsrf parameter in POST requests. LTI requests are
        meant to be cross-site, so it must not be verified.
        """
        return

    async def get(self):
        """Overrides the upstream get handler and always raise HTTPError 405."""
        raise HTTPError(405, "GET method is not allowed for launch requests")

    async def post(self):
        """
        Overrides the upstream post handler.
        """
        try:
            id_token = self.decode_and_validate_launch_request()
        except InvalidAudienceError as e:
            raise HTTPError(401, reason=str(e))
        except ValidationError as e:
            raise HTTPError(400, reason=str(e))

        try:
            user = await self.login_user(id_token)
        except LoginError as e:
            raise HTTPError(400, reason=str(e))
        self.log.debug(f"user logged in: {user}")
        if user is None:
            raise HTTPError(403, reason="User missing or null")
        await self.redirect_to_next_url(user)

    async def redirect_to_next_url(self, user):
        """Redirect user agent to next url that has been received in the login initiation request."""
        next_url = self.get_next_url(user)
        self.redirect(next_url)
        self.log.debug(f"Redirecting user {user.name} to {next_url}")

    def decode_and_validate_launch_request(self) -> Dict[str, Any]:
        """Decrypt, verify and validate launch request parameters.

        Raises subclasses of `ValidationError` of `HTTPError` if anything fails.

        References:
        https://openid.net/specs/openid-connect-core-1_0.html#IDToken
        https://openid.net/specs/openid-connect-core-1_0.html#ImplicitIDTValidation
        """
        validator = LTI13LaunchValidator()

        args = convert_request_to_dict(self.request.arguments)
        self.log.debug(f"Initial launch request args are {args}")

        validator.validate_auth_response(args)

        # Check is state is the same as in the authorization request issued
        # constructed in `LTI13LoginInitHandler.post`, prevents CSRF
        self.check_state()

        id_token = validator.verify_and_decode_jwt(
            encoded_jwt=args.get("id_token"),
            issuer=self.authenticator.issuer,
            audience=self.authenticator.client_id,
            jwks_endpoint=self.authenticator.jwks_endpoint,
            jwks_algorithms=self.authenticator.jwks_algorithms,
        )
        validator.validate_id_token(id_token)
        validator.validate_azp_claim(id_token, self.authenticator.client_id)

        # Check nonce matches the one that has been used in the authorization request.
        # A nonce is a hash of random state which is stored in a session cookie before
        # redirecting to make authorization request. This mitigates replay attacks.
        #
        # References:
        # https://openid.net/specs/openid-connect-core-1_0.html#NonceNotes
        # https://auth0.com/docs/get-started/authentication-and-authorization-flow/mitigate-replay-attacks-when-using-the-implicit-flow
        self.check_nonce(id_token)

        return id_token

    def check_state(self):
        """Verify OAuth state

        compare value in cookie with redirect url param
        """
        cookie_state = self._get_state_cookie()
        if not cookie_state:
            raise HTTPError(400, reason="OAuth state missing from cookies")
        url_state = self._get_state_from_url()
        if cookie_state != url_state:
            self.log.warning(f"OAuth state mismatch: {cookie_state} != {url_state}")
            raise HTTPError(400, reason="OAuth state mismatch")

    def check_nonce(self, id_token: Dict[str, Any]) -> None:
        """Check if received nonce corresponds to hash of nonce state cookie"""
        received_nonce = id_token.get("nonce")

        nonce_state = self._get_nonce_state_cookie()
        if not nonce_state:
            raise HTTPError(400, reason="Missing nonce state cookie")

        nonce = get_nonce(nonce_state)

        if nonce != received_nonce:
            self.log.warning("OAuth nonce mismatch: %s != %s", nonce, received_nonce)
            raise HTTPError(400, reason="OAuth nonce mismatch")

    def get_next_url(self, user=None):
        """Get the redirect target from the state field"""
        state = self._get_state_from_url()
        next_url = _deserialize_state(state).get("next_url")
        if next_url:
            return next_url
        # JupyterHub 0.8 adds default .get_next_url for a fallback
        return super().get_next_url(user)

    def _get_state_from_url(self):
        """Get OAuth state from URL parameters

        Raises HTTPError(400) if `state` argument is missing from request.
        """
        try:
            return self.get_argument("state")
        except MissingArgumentError:
            raise HTTPError(400, reason="OAuth state missing from URL")

    def _get_nonce_state_cookie(self):
        """Get OAuth nonce state from cookies

        To be compared with the value in id_token
        """
        if self._nonce_state_cookie is None:
            self._nonce_state_cookie = self._get_oauth_cookie(NONCE_STATE_COOKIE_NAME)
        return self._nonce_state_cookie

    def _get_state_cookie(self):
        """Get OAuth state from cookies

        To be compared with the value in redirect URL
        """
        if self._state_cookie is None:
            self._state_cookie = self._get_oauth_cookie(STATE_COOKIE_NAME)
        return self._state_cookie

    def _get_oauth_cookie(self, name: str):
        """Get OAuth state cookie."""
        cookie = (self.get_secure_cookie(name) or b"").decode("utf8", "replace")
        self.clear_cookie(name)
        return cookie



import traitlets.config

from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.handlers.base_handler import GraderBaseHandler, authorize
from grader_service.orm.takepart import Scope
from grader_service.registry import VersionSpecifier, register_handler


def _get_effective_executor_value(
    app_cfg: traitlets.config.Config, executor_class: type[LocalAutogradeExecutor], trait_name: str
):
    """
    Return the configured value for trait_name if present in app_cfg,
    otherwise return the class default pulled from the trait metadata.
    """

    # 1) retrieve the provided class field from the config
    executor_class_field = app_cfg.get(executor_class.__name__, None)

    # executor_class_field may be None, or a Config object / dict-like. Use mapping access.
    if executor_class_field is not None and trait_name in executor_class_field:
        return executor_class_field.get(trait_name)

    # 2) fallback to the trait's default from the class metadata
    return executor_class.class_traits()[trait_name].default()


@register_handler(path=r"\/api\/config\/?", version_specifier=VersionSpecifier.ALL)
class ConfigHandler(GraderBaseHandler):
    """
    Handler class for requests to /config
    """

    @authorize([Scope.tutor, Scope.instructor])
    async def get(self):
        app_cfg = self.application.config
        executor_class = app_cfg.RequestHandlerConfig.autograde_executor_class

        def resolve(name):
            return _get_effective_executor_value(app_cfg, executor_class, name)

        self.write_json(
            {
                "default_cell_timeout": resolve("default_cell_timeout"),
                "min_cell_timeout": resolve("min_cell_timeout"),
                "max_cell_timeout": resolve("max_cell_timeout"),
            }
        )



import typing as t
from textwrap import dedent
from typing import Any, Optional, Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbconvert.preprocessors import ExecutePreprocessor
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Integer, List

from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class UnresponsiveKernelError(Exception):
    pass


class Execute(NbGraderPreprocessor, ExecutePreprocessor):
    interrupt_on_timeout = Bool(True)
    allow_errors = Bool(True)
    raise_on_iopub_timeout = Bool(True)
    extra_arguments = List(
        [],
        help=dedent(
            """
        A list of extra arguments to pass to the kernel. For python kernels,
        this defaults to ``--HistoryManager.hist_file=:memory:``. For other
        kernels this is just an empty list.
        """
        ),
    ).tag(config=True)

    execute_retries = Integer(
        0,
        help=dedent(
            """
        The number of times to try re-executing the notebook before throwing
        an error. Generally, this shouldn't need to be set, but might be useful
        for CI environments when tests are flaky.
        """
        ),
    ).tag(config=True)

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict, retries: Optional[Any] = None
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # This gets added in by the parent execute preprocessor, so if it's already in our
        # extra arguments we need to delete it or traitlets will be unhappy.
        if "--HistoryManager.hist_file=:memory:" in self.extra_arguments:
            self.extra_arguments.remove("--HistoryManager.hist_file=:memory:")

        if retries is None:
            retries = self.execute_retries
        try:
            output = super(Execute, self).preprocess(nb, resources)
        except RuntimeError:
            if retries == 0:
                raise UnresponsiveKernelError()
            else:
                self.log.warning("Failed to execute notebook, trying again...")
                return self.preprocess(nb, resources, retries=retries - 1)

        return output

    async def _async_handle_timeout(
        self, timeout: int, cell: t.Optional[NotebookNode] = None
    ) -> None:
        await super()._async_handle_timeout(timeout, cell)

        error_output = NotebookNode(output_type="error")
        error_output.ename = "CellTimeoutError"
        error_output.evalue = "CellTimeoutError"
        error_output.traceback = [
            f"CellTimeoutInterrupt: Cell execution timed out after maximum execution time of {self.timeout} seconds."
        ]
        cell.outputs.append(error_output)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# import handlers
from grader_service.handlers import (
    assignment,
    base_handler,
    config,
    git,
    grading,
    health,
    lectures,
    permission,
    submissions,
)
from grader_service.handlers.handler_utils import GitRepoType

__all__ = [
    "assignment",
    "grading",
    "lectures",
    "submissions",
    "git",
    "permission",
    "health",
    "config",
    "base_handler",
    "GitRepoType",
]



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from http import HTTPStatus

import celery

from grader_service.autograding.celery.tasks import (
    autograde_task,
    generate_feedback_task,
    lti_sync_task,
)
from grader_service.handlers.base_handler import GraderBaseHandler, authorize
from grader_service.orm.submission import AutoStatus, FeedbackStatus, Submission
from grader_service.orm.takepart import Scope
from grader_service.registry import VersionSpecifier, register_handler

from .handler_utils import parse_ids


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments"
    + r"\/(?P<assignment_id>\d*)\/grading\/(?P<sub_id>\d*)\/auto\/?",
    version_specifier=VersionSpecifier.ALL,
)
class GradingAutoHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/grading
    /{submission_id}/auto.
    """

    def on_finish(self):
        # we do not close the session we just commit
        # because LocalAutogradeExecutor still needs it
        if self.session:
            self.session.commit()

    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int, sub_id: int):
        """
        Starts the autograding process of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param sub_id: id of the submission
        :type sub_id: int
        :raises HTTPError: throws err if the submission was not found
        """
        lecture_id, assignment_id, sub_id = parse_ids(lecture_id, assignment_id, sub_id)
        self.validate_parameters()
        submission = self.get_submission(lecture_id, assignment_id, sub_id)
        submission.auto_status = AutoStatus.PENDING
        if submission.feedback_status == FeedbackStatus.GENERATED:
            submission.feedback_status = FeedbackStatus.FEEDBACK_OUTDATED
        self.session.commit()

        submission = self.session.get(Submission, sub_id)

        autograde_task.delay(lecture_id, assignment_id, sub_id)
        self.set_status(HTTPStatus.ACCEPTED, reason="Autograding submission process started")

        self.write_json(submission)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments"
    + r"\/(?P<assignment_id>\d*)\/grading\/(?P<sub_id>\d*)\/feedback\/?",
    version_specifier=VersionSpecifier.ALL,
)
class GenerateFeedbackHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/grading/
    {submission_id}/feedback.
    """

    def on_finish(self):
        # we do not close the session we just commit
        # because GenerateFeedbackHandler still needs it
        if self.session:
            self.session.commit()

    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int, sub_id):
        """
        Starts the process of generating feedback for a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param sub_id: id of the submission
        :type sub_id: int
        :raises HTTPError: throws err if the submission was not found
        """
        lecture_id, assignment_id, sub_id = parse_ids(lecture_id, assignment_id, sub_id)
        self.validate_parameters()
        lecture = self.get_lecture(lecture_id)
        assignment = self.get_assignment(lecture_id, assignment_id)
        submission: Submission = self.get_submission(lecture_id, assignment_id, sub_id)
        submission.feedback_status = FeedbackStatus.GENERATING
        self.session.commit()
        # use immutable signature: https://docs.celeryq.dev/en/stable/reference/celery.app.task.html#celery.app.task.Task.si
        generate_feedback_chain = celery.chain(
            generate_feedback_task.si(lecture_id, assignment_id, sub_id),
            lti_sync_task.si(
                lecture.serialize(),
                assignment.serialize(),
                [submission.serialize_with_user()],
                feedback_sync=True,
            ),
        )
        generate_feedback_chain()

        self.set_status(HTTPStatus.ACCEPTED, reason="Generating feedback process started")
        self.write_json(submission)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import os
import subprocess
from typing import Any, Set

from traitlets.traitlets import Unicode

from grader_service.autograding.git_manager import GitSubmissionManager
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.convert.converters.generate_feedback import GenerateFeedback
from grader_service.handlers.handler_utils import GitRepoType
from grader_service.orm.submission import FeedbackStatus, Submission


class FeedbackGitSubmissionManager(GitSubmissionManager):
    """Git manager for generating submission feedback."""

    input_repo_type = GitRepoType.AUTOGRADE
    output_repo_type = GitRepoType.FEEDBACK

    def __init__(self, grader_service_dir: str, submission: Submission, **kwargs: Any):
        super().__init__(grader_service_dir, submission, **kwargs)
        self.input_branch = f"submission_{self.submission.commit_hash}"
        self.output_branch = f"feedback_{self.submission.commit_hash}"


class LocalFeedbackExecutor(LocalAutogradeExecutor):
    git_manager_class = FeedbackGitSubmissionManager

    @property
    def input_path(self):
        return os.path.join(
            self.grader_service_dir, self.relative_input_path, f"feedback_{self.submission.id}"
        )

    @property
    def output_path(self):
        return os.path.join(
            self.grader_service_dir, self.relative_output_path, f"feedback_{self.submission.id}"
        )

    def _run(self):
        feedback_generator = GenerateFeedback(
            self.input_path,
            self.output_path,
            "*.ipynb",
            assignment_settings=self.assignment.settings,
        )
        feedback_generator.start()

    def _put_grades_in_assignment_properties(self) -> str:
        # No need to calculate the properties again when generating feedback.
        return self.submission.properties.properties

    def _get_whitelist_patterns(self) -> Set[str]:
        # We only want to commit html files when generating feedback.
        return {"*.html"}

    def _set_properties(self) -> None:
        # No need to set properties.
        pass

    def _set_db_state(self, success=True) -> None:
        """
        Sets the submission feedback status based on the success of the generation.

        :param success: Whether feedback generation was successful or not.
        :return: None
        """
        if success:
            self.submission.feedback_status = FeedbackStatus.GENERATED
        else:
            self.submission.feedback_status = FeedbackStatus.GENERATION_FAILED
        self.session.commit()

    def _update_submission_logs(self):
        # No need to collect feedback logs.
        pass


class LocalFeedbackProcessExecutor(LocalFeedbackExecutor):
    convert_executable = Unicode("grader-convert", allow_none=False).tag(config=True)

    def _run(self):
        command = [
            self.convert_executable,
            "generate_feedback",
            "-i",
            self.input_path,
            "-o",
            self.output_path,
            "-p",
            "*.ipynb",
        ]
        self.log.info(f"Running {command}")
        process = subprocess.run(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=None, text=True
        )
        if process.returncode == 0:
            self.log.info(process.stderr)
            self.log.info("Process has successfully completed execution!")
        else:
            self.log.error(process.stderr)
            raise RuntimeError("Process has failed execution!")



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class FeedbackStatus(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    """
    allowed enum values
    """
    NOT_GENERATED = 'not_generated'
    GENERATING = 'generating'
    GENERATED = 'generated'
    GENERATION_FAILED = 'generation_failed'
    FEEDBACK_OUTDATED = 'feedback_outdated'
    def __init__(self):  # noqa: E501
        """FeedbackStatus - a model defined in OpenAPI

        """
        self.openapi_types = {
        }

        self.attribute_map = {
        }

    @classmethod
    def from_dict(cls, dikt) -> 'FeedbackStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The FeedbackStatus of this FeedbackStatus.  # noqa: E501
        :rtype: FeedbackStatus
        """
        return util.deserialize_model(dikt, cls)



"""Settings column

Revision ID: 4620c13728e7
Revises: a1791b1371ed
Create Date: 2023-09-11 15:46:57.162396

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4620c13728e7"
down_revision = "a1791b1371ed"
branch_labels = None
depends_on = None


def upgrade():
    # use settings to save arbitrary json
    op.add_column("assignment", sa.Column("settings", sa.Text, server_default="{}", nullable=False))

    # columns for original points (grading_points) and the late submission scaling (point_scaling)
    op.add_column("submission", sa.Column("grading_score", sa.DECIMAL(10, 3), nullable=True))
    op.add_column(
        "submission",
        sa.Column("score_scaling", sa.DECIMAL(10, 3), server_default="1.0", nullable=False),
    )


def downgrade():
    op.drop_column("assignment", "settings")
    op.drop_column("submission", "grading_score")
    op.drop_column("submission", "score_scaling")



from typing import Dict, Tuple

from nbformat.notebooknode import NotebookNode

from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class AddRevert(NbGraderPreprocessor):
    """Adds original code to cell which is used to revert the cell in
    extension"""

    def preprocess(self, nb: NotebookNode, resources: Dict) -> Tuple[NotebookNode, Dict]:
        for cell in nb.cells:
            if "nbgrader" not in cell.metadata:
                cell.metadata["revert"] = cell.source
                continue

            if cell.metadata is None or cell.metadata["nbgrader"]["solution"]:
                cell.metadata["revert"] = cell.source

        return nb, resources



from textwrap import dedent
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Unicode

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ClearMarkScheme(NbGraderPreprocessor):
    begin_mark_scheme_delimeter = Unicode(
        "BEGIN MARK SCHEME", help="The delimiter marking the beginning of hidden tests cases"
    ).tag(config=True)

    end_mark_scheme_delimeter = Unicode(
        "END MARK SCHEME", help="The delimiter marking the end of hidden tests cases"
    ).tag(config=True)

    enforce_metadata = Bool(
        True,
        help=dedent(
            """
            Whether or not to complain if cells containing marking scheme regions
            are not marked as task cells. WARNING: this will potentially cause
            things to break if you are using the full nbgrader pipeline. ONLY
            disable this option if you are only ever planning to use nbgrader
            assign.
            """
        ),
    ).tag(config=True)

    def _remove_mark_scheme_region(self, cell: NotebookNode) -> bool:
        """Find a region in the cell that is delimeted by
        `self.begin_mark_scheme_delimeter` and `self.end_mark_scheme_delimeter` (e.g.  ###
        BEGIN MARK SCHEME and ### END MARK SCHEME). Remove that region
        depending the cell type.

        This modifies the cell in place, and then returns True if a
        mark region was removed, and False otherwise.
        """
        # pull out the cell input/source
        lines = cell.source.split("\n")

        new_lines = []
        in_ms = False
        removed_ms = False

        for line in lines:
            # begin the test area
            if self.begin_mark_scheme_delimeter in line:
                # check to make sure this isn't a nested BEGIN HIDDEN TESTS
                # region
                if in_ms:
                    raise RuntimeError("Encountered nested begin mark scheme statements")
                in_ms = True
                removed_ms = True

            # end the solution area
            elif self.end_mark_scheme_delimeter in line:
                in_ms = False

            # add lines as long as it's not in the hidden tests region
            elif not in_ms:
                new_lines.append(line)

        # we finished going through all the lines, but didn't find a
        # matching END HIDDEN TESTS statment
        if in_ms:
            raise RuntimeError("No end mark scheme tests statement found")

        # replace the cell source
        cell.source = "\n".join(new_lines)

        return removed_ms

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        nb, resources = super(ClearMarkScheme, self).preprocess(nb, resources)
        if "celltoolbar" in nb.metadata:
            del nb.metadata["celltoolbar"]
        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # remove hidden test regions
        removed_ms = self._remove_mark_scheme_region(cell)

        # determine whether the cell is a task cell
        is_task = utils.is_task(cell)

        # check that it is marked as a task cell if we remove a mark scheme
        # region -- if it's not, then this is a problem, because the cell needs
        # to be given an id
        if not is_task and removed_ms:
            if self.enforce_metadata:
                raise RuntimeError(
                    "Mark scheme region detected in a non-grade cell; "
                    "please make sure all mark scheme regions are within "
                    "'Manually graded task' cells."
                )

        return cell, resources



from nbconvert.preprocessors import ClearOutputPreprocessor

from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ClearOutput(NbGraderPreprocessor, ClearOutputPreprocessor):
    pass



"""merged assignment configuration options into assignment settings column

Revision ID: fc5d2febe781
Revises: a0718dae969d
Create Date: 2025-01-27 15:42:24.363658

"""

import datetime
import json

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fc5d2febe781"
down_revision = "a0718dae969d"
branch_labels = None
depends_on = None


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def upgrade():
    # Migrate data from old columns to the settings column
    conn = op.get_bind()
    assignments = conn.execute(
        sa.text(
            "SELECT id, duedate, type, automatic_grading, max_submissions, allow_files, settings FROM assignment"
        )
    )

    for assignment in assignments:
        assignment = dict(zip(assignments.keys(), assignment))
        existing_settings = json.loads(assignment["settings"]) if assignment["settings"] else {}
        new_settings = {
            "deadline": assignment["duedate"] if assignment["duedate"] else None,
            "assignment_type": assignment["type"],
            "autograde_type": assignment["automatic_grading"],
            "max_submissions": assignment["max_submissions"],
            "allowed_files": ["*"] if bool(assignment["allow_files"]) else [],
        }
        existing_settings.update(new_settings)
        conn.execute(
            sa.text("UPDATE assignment SET settings = :settings WHERE id = :id"),
            {
                "settings": json.dumps(existing_settings, default=json_serial),
                "id": assignment["id"],
            },
        )

    # Drop the old columns
    op.drop_column("assignment", "duedate")
    op.drop_column("assignment", "type")
    op.drop_column("assignment", "automatic_grading")
    op.drop_column("assignment", "max_submissions")
    op.drop_column("assignment", "allow_files")


def downgrade():
    # Recreate the old columns
    # Drop the old columns
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    columns = [col["name"] for col in inspector.get_columns("assignment")]
    if "duedate" not in columns:
        op.add_column("assignment", sa.Column("duedate", sa.DateTime(), nullable=True))
    if "type" not in columns:
        op.add_column(
            "assignment",
            sa.Column(
                "type",
                sa.Enum("user", "group", name="assignment_type"),
                nullable=False,
                server_default="user",
            ),
        )
    if "automatic_grading" not in columns:
        op.add_column(
            "assignment",
            sa.Column(
                "automatic_grading",
                sa.Enum("unassisted", "auto", "full_auto", name="automatic_grading"),
                server_default="unassisted",
                nullable=False,
            ),
        )
    if "max_submissions" not in columns:
        op.add_column(
            "assignment",
            sa.Column(
                "max_submissions", sa.Integer(), nullable=True, server_default=None, unique=False
            ),
        )
    if "allow_files" not in columns:
        op.add_column(
            "assignment",
            sa.Column(
                "allow_files", sa.Boolean(), nullable=False, server_default="f", default=False
            ),
        )

    # Migrate data back from the settings column to the old columns
    assignments = conn.execute(sa.text("SELECT id, settings FROM assignment"))

    for assignment in assignments:
        assignment = dict(zip(assignments.keys(), assignment))
        try:
            settings = json.loads(assignment["settings"]) if assignment["settings"] else {}
        except json.JSONDecodeError:
            settings = {}
        conn.execute(
            sa.text(
                """
                UPDATE assignment
                SET duedate = :duedate, type = :type, automatic_grading = :automatic_grading,
                max_submissions = :max_submissions, allow_files = :allow_files
                WHERE id = :id
                """
            ),
            {
                "duedate": (
                    datetime.datetime.fromisoformat(settings["deadline"])
                    if settings.get("deadline")
                    else None
                ),
                "type": settings.get("assignment_type", "user"),
                "automatic_grading": settings.get("autograde_type", "unassisted"),
                "max_submissions": settings.get("max_submissions", None),
                "allow_files": len(settings.get("allowed_files", [])) > 0,
                "id": assignment["id"],
            },
        )



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode

from grader_service.convert import utils
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class ComputeChecksums(NbGraderPreprocessor):
    """A preprocessor to compute checksums of grade cells."""

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # compute checksums of grade cell and solution cells
        if utils.is_grade(cell) or utils.is_solution(cell) or utils.is_locked(cell):
            checksum = utils.compute_checksum(cell)
            cell.metadata.nbgrader["checksum"] = checksum
            cell.metadata.nbgrader["cell_type"] = cell.cell_type

            if utils.is_grade(cell) or utils.is_solution(cell):
                self.log.debug(
                    "Checksum for %s cell '%s' is %s",
                    cell.metadata.nbgrader["cell_type"],
                    cell.metadata.nbgrader["grade_id"],
                    checksum,
                )

        return cell, resources



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import os

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PrefixLoader
from tornado import web
from traitlets import Dict, Float, List, Unicode, config, default

GRADER_COOKIE_NAME = "grader_service_login"


class GraderServer(config.LoggingConfigurable, web.Application):
    """As an unmanage jupyter hub service, the application gets these
    environment variables from the hub

    See jupyterhub docs section on launching-a-hub-managed-service:
    https://jupyterhub.readthedocs.io/en/stable/reference/services.html#launching-a-hub-managed-service
    """

    cookie_max_age_days = Float(
        14,
        help="""Number of days for a login cookie to be valid.
        Default is two weeks.
        """,
    ).tag(config=True)

    static_file_path = Unicode(
        os.path.join(os.path.dirname(__file__), "static"),
        help="""Absolute path to the static file directory.""",
    ).tag(config=True)

    logo_file = Unicode(
        "", help="Specify path to a logo image to override the Grader logo in the banner."
    ).tag(config=True)

    @default("logo_file")
    def _logo_file_default(self):
        return os.path.join(os.path.dirname(__file__), "static", "images", "grader_logo.png")

    template_paths = List(
        help="Paths to search for jinja templates, before using the default templates."
    ).tag(config=True)

    @default("template_paths")
    def _template_paths_default(self):
        return [os.path.join(os.path.dirname(__file__), "templates")]

    jinja_environment_options = Dict(
        help="Supply extra arguments that will be passed to Jinja environment."
    ).tag(config=True)

    template_vars = Dict(
        help="""Extra variables to be passed into jinja templates.

        Values in dict may contain callable objects.
        If value is callable, the current user is passed as argument.

        Example::

            def callable_value(user):
                # user is generated by handlers.base.get_current_user
                with open("/tmp/file.txt", "r") as f:
                    ret = f.read()
                ret = ret.replace("<username>", user.name)
                return ret

            c.JupyterHub.template_vars = {
                "key1": "value1",
                "key2": callable_value,
            }
        """
    ).tag(config=True)

    def __init__(
        self,
        grader_service_dir: str,
        base_url: str,
        authenticator,
        oauth_provider,
        session_maker,
        **kwargs,
    ):
        kwargs.update(dict(static_path=self.static_file_path))
        super().__init__(**kwargs)
        self.grader_service_dir = grader_service_dir
        self.base_url = base_url
        self.authenticator = authenticator
        self.oauth_provider = oauth_provider
        self.cookie_name = GRADER_COOKIE_NAME
        self.session_maker = session_maker

        jinja_options = dict(autoescape=True, enable_async=True)
        jinja_options.update(self.jinja_environment_options)
        base_path = self._template_paths_default()[0]
        if base_path not in self.template_paths:
            self.template_paths.append(base_path)
        loader = ChoiceLoader(
            [
                PrefixLoader({"templates": FileSystemLoader([base_path])}, "/"),
                FileSystemLoader(self.template_paths),
            ]
        )
        self.jinja_env = Environment(loader=loader, **jinja_options)
        # We need a sync jinja environment too, for the times we *must* use sync
        # code - particularly in RequestHandler.write_error. Since *that*
        # is called from inside the asyncio event loop, we can't actulaly just
        # schedule it on the loop - without starting another thread with its
        # own loop, which seems not worth the trouble. Instead, we create another
        # environment, exactly like this one, but sync
        del jinja_options["enable_async"]
        self.jinja_env_sync = Environment(loader=loader, **jinja_options)



from .common import SchemaTooNewError, SchemaTooOldError, ValidationError
from .v3 import MetadataValidatorV3 as MetadataValidator
from .v3 import read_v3 as read
from .v3 import reads_v3 as reads
from .v3 import write_v3 as write
from .v3 import writes_v3 as writes

SCHEMA_VERSION = MetadataValidator.schema_version

__all__ = [
    "SchemaTooNewError",
    "SchemaTooOldError",
    "ValidationError",
    "MetadataValidator",
    "read",
    "reads",
    "write",
    "writes",
]



import os
from typing import Any

from nbconvert.exporters import HTMLExporter
from nbconvert.preprocessors import CSSHTMLHeaderPreprocessor
from traitlets import List, default
from traitlets.config import Config

from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.convert import utils
from grader_service.convert.converters.base import BaseConverter
from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.gradebook.gradebook import Gradebook
from grader_service.convert.preprocessors import GetGrades


class GenerateFeedback(BaseConverter):
    preprocessors = List([GetGrades, CSSHTMLHeaderPreprocessor]).tag(config=True)

    @default("classes")
    def _classes_default(self):
        classes = super(GenerateFeedback, self)._classes_default()
        classes.append(HTMLExporter)
        return classes

    @default("export_class")
    def _exporter_class_default(self):
        return HTMLExporter

    @default("permissions")
    def _permissions_default(self):
        return 664

    def __init__(
        self,
        input_dir: str,
        output_dir: str,
        file_pattern: str,
        assignment_settings: AssignmentSettings,
        **kwargs: Any,
    ):
        super(GenerateFeedback, self).__init__(
            input_dir, output_dir, file_pattern, assignment_settings, **kwargs
        )
        c = Config()
        # Note: nbconvert 6.0 completely changed how templates work: they can now be installed separately
        #  and can be given by name (classic is default)
        if "template" not in self.config.HTMLExporter:
            template_path = os.path.abspath(
                os.path.join(os.path.dirname(__file__), "..", "templates")
            )
            # '/Users/matthiasmatt/opt/miniconda3/envs/grader/share/jupyter/nbconvert/templates/classic/index.html.j2'
            c.TemplateExporter.extra_template_basedirs = template_path
            c.HTMLExporter.template_name = "feedback"
        self.update_config(c)
        self.force = True  # always overwrite generated assignments

    def init_notebooks(self) -> None:
        super().init_notebooks()
        json_path = os.path.join(self._output_directory, "gradebook.json")
        with Gradebook(json_path) as gb:
            # `self.notebooks` contains the notebooks actually submitted by the student.
            # `notebook_id_set` contains the original notebooks from the assignment.
            # We generate feedback for the notebooks belonging to these both sets.
            student_nbs = {
                self.init_single_notebook_resources(nb)["unique_key"]: nb for nb in self.notebooks
            }
            assign_nb_ids = gb.model.notebook_id_set
            self.notebooks = [path for id, path in student_nbs.items() if id in assign_nb_ids]

        if len(self.notebooks) == 0:
            self.log.warning("No notebooks to generate feedback")

    def get_include_patterns(self, gb: Gradebook) -> list[str]:
        """Get glob patterns specifying for which submission files to generate feedback.

        In case of feedback generation, it can only be done for the notebooks
        which were included in the original assignment.
        """
        return self.notebooks


class GenerateFeedbackApp(ConverterApp):
    version = ConverterApp.__version__

    def start(self):
        GenerateFeedback(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            assignment_settings=utils.get_assignment_settings_from_env(),
            config=self.config,
        ).start()



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.api import util

from grader_service.api.models.assignment_settings import AssignmentSettings  # noqa: E501

class Assignment(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, id=None, name=None, status=None, points=None, settings=None):  # noqa: E501
        """Assignment - a model defined in OpenAPI

        :param id: The id of this Assignment.  # noqa: E501
        :type id: int
        :param name: The name of this Assignment.  # noqa: E501
        :type name: str
        :param status: The status of this Assignment.  # noqa: E501
        :type status: str
        :param points: The points of this Assignment.  # noqa: E501
        :type points: float
        :param settings: The settings of this Assignment.  # noqa: E501
        :type settings: AssignmentSettings
        """
        self.openapi_types = {
            'id': int,
            'name': str,
            'status': str,
            'points': float,
            'settings': AssignmentSettings
        }

        self.attribute_map = {
            'id': 'id',
            'name': 'name',
            'status': 'status',
            'points': 'points',
            'settings': 'settings'
        }

        self._id = id
        self._name = name
        self._status = status
        self._points = points
        self._settings = settings

    @classmethod
    def from_dict(cls, dikt) -> 'Assignment':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Assignment of this Assignment.  # noqa: E501
        :rtype: Assignment
        """
        return util.deserialize_model(dikt, cls)

    @property
    def id(self) -> int:
        """Gets the id of this Assignment.


        :return: The id of this Assignment.
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id: int):
        """Sets the id of this Assignment.


        :param id: The id of this Assignment.
        :type id: int
        """

        self._id = id

    @property
    def name(self) -> str:
        """Gets the name of this Assignment.


        :return: The name of this Assignment.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name of this Assignment.


        :param name: The name of this Assignment.
        :type name: str
        """

        self._name = name

    @property
    def status(self) -> str:
        """Gets the status of this Assignment.


        :return: The status of this Assignment.
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status: str):
        """Sets the status of this Assignment.


        :param status: The status of this Assignment.
        :type status: str
        """
        allowed_values = ["created", "pushed", "released", "complete"]  # noqa: E501
        if status not in allowed_values:
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"
                .format(status, allowed_values)
            )

        self._status = status

    @property
    def points(self) -> float:
        """Gets the points of this Assignment.


        :return: The points of this Assignment.
        :rtype: float
        """
        return self._points

    @points.setter
    def points(self, points: float):
        """Sets the points of this Assignment.


        :param points: The points of this Assignment.
        :type points: float
        """

        self._points = points

    @property
    def settings(self) -> AssignmentSettings:
        """Gets the settings of this Assignment.


        :return: The settings of this Assignment.
        :rtype: AssignmentSettings
        """
        return self._settings

    @settings.setter
    def settings(self, settings: AssignmentSettings):
        """Sets the settings of this Assignment.


        :param settings: The settings of this Assignment.
        :type settings: AssignmentSettings
        """

        self._settings = settings



"""add ondelete cascade

Revision ID: 4a88dacd888f
Revises: 9983ef1fda76
Create Date: 2025-11-04 11:55:10.513853

"""

from typing import Dict, List

from alembic import op
from sqlalchemy import (
    Column,
    Connection,
    Inspector,
    MetaData,
    String,
    Table,
    and_,
    inspect,
    select,
    text,
)

# revision identifiers, used by Alembic.
revision = "4a88dacd888f"
down_revision = "9983ef1fda76"
branch_labels = None
depends_on = None


"""
- Standardizes the names of all foreign keys regardless of the database used.
    - SQLite: None -> fk_<table>_<local_col>
    - PostgreSQL: <table>_<local_col>_fkey -> fk_<table>_<local_col>
- Adds the new foreign key “fk_api_token_client_id”.
- Due to the new foreign key “fk_api_token_client_id,” invalid “api_token” entries are deleted.
- All foreign keys receive the option ondelete="CASCADE".
- The new table “alembic_fk_metadata” stores the old foreign key names so that they can be restored during downgrade.
The exception is SQLite, because SQLAlchemy does not allow “None” in names, the new foreign key names are retained.
"""


def _drop_all_foreign_keys(batch_op, connection, table_name: str):
    """
    Remove all foreign key constraints from the specified table.

    Only drops constraints that have a defined name, since unnamed foreign keys
    cannot be dropped explicitly. (SQLite)
    """
    inspector = Inspector.from_engine(connection)
    fks = inspector.get_foreign_keys(table_name)
    for fk in fks:
        if fk["name"] is not None:
            batch_op.drop_constraint(fk["name"], type_="foreignkey")


def _get_fk_name(conn, table_name, local_cols, referred_table):
    """Return the foreign key name if exists, else None (SQLite may be None)"""
    inspector = inspect(conn)
    for fk in inspector.get_foreign_keys(table_name):
        if fk["constrained_columns"] == local_cols and fk["referred_table"] == referred_table:
            return fk["name"]
    return None


def _store_old_fk(conn, table_name, local_cols, referred_table, old_name):
    """Store information about an existing foreign key in a helper table."""
    metadata = MetaData()
    fk_table = Table(
        "alembic_fk_metadata",
        metadata,
        Column("table_name", String, primary_key=True),
        Column("column_name", String, primary_key=True),
        Column("referred_table", String),
        Column("old_fk_name", String),
    )
    fk_table.create(conn, checkfirst=True)
    conn.execute(
        fk_table.insert().values(
            table_name=table_name,
            column_name=local_cols[0],
            referred_table=referred_table,
            old_fk_name=old_name,
        )
    )


def _get_stored_old_fk(conn, table_name, local_cols, referred_table):
    """Retrieve old foreign key name from a helper table"""
    metadata = MetaData()
    fk_table = Table("alembic_fk_metadata", metadata, autoload_with=conn)
    stmt = select(fk_table.c.old_fk_name).where(
        and_(
            fk_table.c.table_name == table_name,
            fk_table.c.column_name == local_cols[0],
            fk_table.c.referred_table == referred_table,
        )
    )
    result = conn.execute(stmt).first()
    return result[0] if result else None


def _upgrade_recreate_foreign_keys(
    connection: Connection, table_name: str, fk_definitions: List[Dict]
) -> None:
    """
    Recreate foreign key constraints on a table during a migration.

    Steps performed:
    1. Retrieve the current foreign key names for the given definitions.
    2. Store the old foreign key info in a helper table for reference.
    3. Drop all existing foreign key constraints on the table.
    4. Recreate the foreign keys using the new definitions with CASCADE on delete.

    :param connection: SQLAlchemy connection object.
    :type connection: sqlalchemy.engine.Connection
    :param table_name: Name of the table to alter.
    :type table_name: str
    :param fk_definitions: List of dictionaries describing foreign keys to recreate, each containing:
        - new_constraint_name: Name for the new foreign key constraint.
        - referred_table: Name of the referenced table.
        - local_cols: List of columns in the local table.
        - remote_cols: List of columns in the referenced table.
    :type fk_definitions: List[Dict]
    """
    with op.batch_alter_table(table_name) as batch_op:
        for fk in fk_definitions:
            old_name = _get_fk_name(connection, table_name, fk["local_cols"], fk["referred_table"])
            _store_old_fk(connection, table_name, fk["local_cols"], fk["referred_table"], old_name)

        _drop_all_foreign_keys(batch_op, connection, table_name)

        for fk in fk_definitions:
            batch_op.create_foreign_key(
                constraint_name=fk["new_constraint_name"],
                referent_table=fk["referred_table"],
                local_cols=fk["local_cols"],
                remote_cols=fk["remote_cols"],
                ondelete="CASCADE",
            )


def _downgrade_recreate_foreign_keys(
    connection: Connection, table_name: str, fk_definitions: List[Dict]
) -> None:
    """
    Restore previous foreign key constraints during a downgrade migration.

    For each foreign key in fk_definitions:
    1. Retrieve the old foreign key name from the helper table.
    2. Drop the new foreign key constraint (except for SQLite, which handles it differently).
    3. Recreate the old foreign key using the stored name, or fallback to the new name if none was stored.

    :param connection: SQLAlchemy connection object.
    :type connection: sqlalchemy.engine.Connection
    :param table_name: Name of the table to alter.
    :type table_name: str
    :param fk_definitions: List of dictionaries describing foreign keys to restore, each containing:
        - new_constraint_name: Name of the current foreign key constraint.
        - referred_table: Name of the referenced table.
        - local_cols: List of columns in the local table.
        - remote_cols: List of columns in the referenced table.
    :type fk_definitions: List[Dict]
    """
    with op.batch_alter_table(table_name) as batch_op:
        for fk in fk_definitions:
            old_fk_name = _get_stored_old_fk(
                connection, table_name, fk["local_cols"], fk["referred_table"]
            )
            if connection.dialect.name != "sqlite":
                batch_op.drop_constraint(fk["new_constraint_name"], type_="foreignkey")
            if old_fk_name is None:
                if connection.dialect.name != "sqlite":
                    return
                old_fk_name = fk["new_constraint_name"]
            batch_op.create_foreign_key(
                constraint_name=old_fk_name,
                referent_table=fk["referred_table"],
                local_cols=fk["local_cols"],
                remote_cols=fk["remote_cols"],
            )


def upgrade():
    connection = op.get_bind()
    dialect = op.get_bind().dialect.name

    if dialect == "sqlite":
        connection.execute(text("PRAGMA foreign_keys=OFF;"))

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="assignment",
        fk_definitions=[
            {
                "new_constraint_name": "fk_assignment_lectid",
                "referred_table": "lecture",
                "local_cols": ["lectid"],
                "remote_cols": ["id"],
            }
        ],
    )

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_assignid",
                "referred_table": "assignment",
                "local_cols": ["assignid"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_submission_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
        ],
    )

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission_logs",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_logs_sub_id",
                "referred_table": "submission",
                "local_cols": ["sub_id"],
                "remote_cols": ["id"],
            }
        ],
    )

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission_properties",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_properties_sub_id",
                "referred_table": "submission",
                "local_cols": ["sub_id"],
                "remote_cols": ["id"],
            }
        ],
    )

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="takepart",
        fk_definitions=[
            {
                "new_constraint_name": "fk_takepart_lectid",
                "referred_table": "lecture",
                "local_cols": ["lectid"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_takepart_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
        ],
    )

    # OAuth/LTI tokens that reference non-existent clients are deleted
    connection.execute(
        text(
            """
             DELETE
             FROM api_token
             WHERE
               client_id IS NOT NULL  -- non-OAuth tokens may not have client_id set
               AND client_id NOT IN (SELECT identifier FROM oauth_client);
             """
        )
    )
    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="api_token",
        fk_definitions=[
            {
                "new_constraint_name": "fk_api_token_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_api_token_client_id",
                "referred_table": "oauth_client",
                "local_cols": ["client_id"],
                "remote_cols": ["identifier"],
            },
        ],
    )

    _upgrade_recreate_foreign_keys(
        connection=connection,
        table_name="oauth_code",
        fk_definitions=[
            {
                "new_constraint_name": "fk_oauth_code_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_oauth_code_client_id",
                "referred_table": "oauth_client",
                "local_cols": ["client_id"],
                "remote_cols": ["identifier"],
            },
        ],
    )

    if dialect == "sqlite":
        connection.execute(text("PRAGMA foreign_keys=ON;"))


def downgrade():
    connection = op.get_bind()
    dialect = op.get_bind().dialect.name

    if dialect == "sqlite":
        connection.execute(text("PRAGMA foreign_keys=OFF;"))

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="assignment",
        fk_definitions=[
            {
                "new_constraint_name": "fk_assignment_lectid",
                "referred_table": "lecture",
                "local_cols": ["lectid"],
                "remote_cols": ["id"],
            }
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_assignid",
                "referred_table": "assignment",
                "local_cols": ["assignid"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_submission_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission_logs",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_logs_sub_id",
                "referred_table": "submission",
                "local_cols": ["sub_id"],
                "remote_cols": ["id"],
            }
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="submission_properties",
        fk_definitions=[
            {
                "new_constraint_name": "fk_submission_properties_sub_id",
                "referred_table": "submission",
                "local_cols": ["sub_id"],
                "remote_cols": ["id"],
            }
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="takepart",
        fk_definitions=[
            {
                "new_constraint_name": "fk_takepart_lectid",
                "referred_table": "lecture",
                "local_cols": ["lectid"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_takepart_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="api_token",
        fk_definitions=[
            {
                "new_constraint_name": "fk_api_token_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_api_token_client_id",
                "referred_table": "oauth_client",
                "local_cols": ["client_id"],
                "remote_cols": ["identifier"],
            },
        ],
    )

    _downgrade_recreate_foreign_keys(
        connection=connection,
        table_name="oauth_code",
        fk_definitions=[
            {
                "new_constraint_name": "fk_oauth_code_user_id",
                "referred_table": "user",
                "local_cols": ["user_id"],
                "remote_cols": ["id"],
            },
            {
                "new_constraint_name": "fk_oauth_code_client_id",
                "referred_table": "oauth_client",
                "local_cols": ["client_id"],
                "remote_cols": ["identifier"],
            },
        ],
    )

    op.drop_table("alembic_fk_metadata")

    if dialect == "sqlite":
        connection.execute(text("PRAGMA foreign_keys=ON;"))



import json
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode

from grader_service.convert import utils
from grader_service.convert.gradebook.gradebook import Gradebook, MissingEntry
from grader_service.convert.gradebook.models import GradeCell, SolutionCell, SourceCell, TaskCell
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class SaveCells(NbGraderPreprocessor):
    """A preprocessor to save information about grade and solution cells."""

    def _create_notebook(self, nb: NotebookNode) -> None:
        notebook_info = None

        try:
            notebook = self.gradebook.find_notebook(self.notebook_id)
        except MissingEntry:
            notebook_info = {}
        else:
            # pull out existing cell ids
            self.old_grade_cells = set(x.name for x in notebook.grade_cells)
            self.old_solution_cells = set(x.name for x in notebook.solution_cells)
            self.old_task_cells = set(x.name for x in notebook.task_cells)
            self.old_source_cells = set(x.name for x in notebook.source_cells)

            # clear data about the existing notebook
            self.log.debug("Removing existing notebook '%s' from the database", self.notebook_id)
            notebook_info = notebook.to_dict()
            del notebook_info["name"]
            del notebook_info["grade_cells_dict"]
            del notebook_info["solution_cells_dict"]
            del notebook_info["task_cells_dict"]
            del notebook_info["source_cells_dict"]
            del notebook_info["grades_dict"]
            del notebook_info["comments_dict"]
            self.gradebook.remove_notebook(self.notebook_id)

        # create the notebook
        if notebook_info is not None:
            kernelspec = nb.metadata.get("kernelspec", {})
            notebook_info["kernelspec"] = json.dumps(kernelspec)
            notebook_info.setdefault("grade_cells_dict", dict())
            notebook_info.setdefault("solution_cells_dict", dict())
            notebook_info.setdefault("task_cells_dict", dict())
            notebook_info.setdefault("source_cells_dict", dict())
            notebook_info.setdefault("grades_dict", dict())
            notebook_info.setdefault("comments_dict", dict())
            notebook_info.setdefault("flagged", False)
            notebook_info["id"] = self.notebook_id
            self.log.debug("Creating notebook '%s' in the database", self.notebook_id)
            self.log.debug("Notebook kernelspec: {}".format(kernelspec))
            self.gradebook.add_notebook(self.notebook_id, **notebook_info)

        # save grade cells
        for name, info in self.new_grade_cells.items():
            grade_cell = self.gradebook.update_or_create_grade_cell(name, self.notebook_id, **info)
            self.log.debug("Recorded grade cell %s into the gradebook", grade_cell)

        # save solution cells
        for name, info in self.new_solution_cells.items():
            solution_cell = self.gradebook.update_or_create_solution_cell(
                name, self.notebook_id, **info
            )
            self.log.debug("Recorded solution cell %s into the gradebook", solution_cell)

        # save task cells
        for name, info in self.new_task_cells.items():
            task_cell = self.gradebook.update_or_create_task_cell(name, self.notebook_id, **info)
            self.log.debug("Recorded task cell %s into the gradebook from info %s", task_cell, info)

        # save source cells
        for name, info in self.new_source_cells.items():
            source_cell = self.gradebook.update_or_create_source_cell(
                name, self.notebook_id, **info
            )
            self.log.debug("Recorded source cell %s into the gradebook", source_cell)

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["unique_key"]
        self.json_path = resources["output_json_path"]

        if self.notebook_id == "":
            raise ValueError("Invalid notebook id: '{}'".format(self.notebook_id))

        # create a place to put new cell information
        self.new_grade_cells = {}
        self.new_solution_cells = {}
        self.new_task_cells = {}
        self.new_source_cells = {}

        # connect to the database
        self.gradebook = Gradebook(self.json_path)

        with self.gradebook:
            nb, resources = super(SaveCells, self).preprocess(nb, resources)

            # create the notebook and save it to the database
            self._create_notebook(nb)

        return nb, resources

    def _create_grade_cell(self, cell: NotebookNode) -> None:
        grade_id = cell.metadata.nbgrader["grade_id"]

        try:
            grade_cell = self.gradebook.find_grade_cell(grade_id, self.notebook_id).to_dict()
            del grade_cell["name"]
        except MissingEntry:
            grade_cell = GradeCell.empty_dict()
            del grade_cell["name"]

        grade_cell.update(
            {"max_score": float(cell.metadata.nbgrader["points"]), "cell_type": cell.cell_type}
        )

        self.new_grade_cells[grade_id] = grade_cell

    def _create_solution_cell(self, cell: NotebookNode) -> None:
        grade_id = cell.metadata.nbgrader["grade_id"]

        try:
            solution_cell = self.gradebook.find_solution_cell(grade_id, self.notebook_id).to_dict()
            del solution_cell["name"]
        except MissingEntry:
            solution_cell = SolutionCell.empty_dict()
            del solution_cell["name"]

        self.new_solution_cells[grade_id] = solution_cell

    def _create_task_cell(self, cell: NotebookNode) -> None:
        grade_id = cell.metadata.nbgrader["grade_id"]
        try:
            task_cell = self.gradebook.find_task_cell(grade_id, self.notebook_id).to_dict()
            del task_cell["name"]
        except MissingEntry:
            task_cell = TaskCell.empty_dict()
            del task_cell["name"]

        task_cell.update(
            {"max_score": float(cell.metadata.nbgrader["points"]), "cell_type": cell.cell_type}
        )

        self.new_task_cells[grade_id] = task_cell

    def _create_source_cell(self, cell: NotebookNode) -> None:
        grade_id = cell.metadata.nbgrader["grade_id"]

        try:
            source_cell = self.gradebook.find_source_cell(grade_id, self.notebook_id).to_dict()
            del source_cell["name"]
        except MissingEntry:
            source_cell = SourceCell.empty_dict()
            del source_cell["name"]

        source_cell.update(
            {
                "cell_type": cell.cell_type,
                "locked": utils.is_locked(cell),
                "source": cell.source,
                "checksum": cell.metadata.nbgrader.get("checksum", None),
            }
        )

        self.new_source_cells[grade_id] = source_cell

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if utils.is_grade(cell):
            self._create_grade_cell(cell)

        if utils.is_solution(cell):
            self._create_solution_cell(cell)

        if utils.is_task(cell):
            self._create_task_cell(cell)

        if (
            utils.is_grade(cell)
            or utils.is_solution(cell)
            or utils.is_locked(cell)
            or utils.is_task(cell)
        ):
            self._create_source_cell(cell)

        return cell, resources



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import asyncio
import inspect
import logging
import os
import secrets
import shutil
import signal
import subprocess
import sys

import tornado
from jupyterhub.log import log_request
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from tornado.httpserver import HTTPServer
from traitlets import (
    Bool,
    Dict,
    Enum,
    HasTraits,
    Instance,
    Int,
    List,
    TraitError,
    Type,
    Unicode,
    config,
    default,
    observe,
    validate,
)
from traitlets import log as traitlets_log

from grader_service import __version__
from grader_service.auth.auth import Authenticator

# run __init__.py to register handlers
from grader_service.auth.dummy import DummyAuthenticator
from grader_service.autograding.celery.app import CeleryApp
from grader_service.handlers.base_handler import RequestHandlerConfig
from grader_service.handlers.static import CacheControlStaticFilesHandler
from grader_service.oauth2 import handlers as oauth_handlers
from grader_service.oauth2.provider import make_provider
from grader_service.orm import Lecture, Role, User
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope
from grader_service.plugins.lti import LTISyncGrades
from grader_service.registry import HandlerPathRegistry
from grader_service.server import GraderServer
from grader_service.utils import url_path_join


def get_session_maker(url) -> scoped_session:
    engine = create_engine(url)
    return scoped_session(sessionmaker(bind=engine))


class GraderService(config.Application):
    name = "grader-service"
    version = __version__

    description = """Starts the grader service, which can be used to create
    and distribute assignments, collect submissions and grade them.
    """

    examples = """
  generate default config file:
      grader-service --generate-config -f /etc/grader/grader_service_config.py
  spawn the grader service:
      grader-service -f /etc/grader/grader_service_config.py
  """

    generate_config = Bool(False, help="Generate config file based on defaults.").tag(config=True)

    service_host = Unicode(
        os.getenv("GRADER_HOST", "0.0.0.0"), help="The host address of the service"
    ).tag(config=True)

    service_port = Int(
        int(os.getenv("GRADER_PORT", "4010")), help="The port the service runs on"
    ).tag(config=True)

    reuse_port = Bool(
        False, help="Whether to allow for the specified service port to be reused."
    ).tag(config=True)

    grader_service_dir = Unicode(os.getenv("GRADER_SERVICE_DIRECTORY"), allow_none=False).tag(
        config=True
    )

    db_url = Unicode(allow_none=False).tag(config=True)

    oauth_provider = None

    @default("db_url")
    def _default_db_url(self):
        db_path = os.path.join(self.grader_service_dir, "grader.db")
        service_dir_url = f"sqlite:///{db_path}"
        return os.getenv("GRADER_DB_URL", service_dir_url)

    grader_cookie_secret = Unicode(
        default_value=os.getenv("GRADER_COOKIE_SECRET", secrets.token_hex(nbytes=32)),
        allow_none=False,
    ).tag(config=True)

    max_body_size = Int(104857600, help="Sets the max body size in bytes, default to 100mb").tag(
        config=True
    )

    max_buffer_size = Int(
        104857600, help="Sets the max buffer size in bytes, default to 100mb"
    ).tag(config=True)

    service_git_username = Unicode("grader-service", allow_none=False).tag(config=True)

    service_git_email = Unicode("", allow_none=False).tag(config=True)

    config_file = Unicode("grader_service_config.py", help="The config file to load").tag(
        config=True
    )

    base_url_path = Unicode("/services/grader/", help="base url path", allow_none=False).tag(
        config=True
    )

    authenticator_class = Type(
        default_value=DummyAuthenticator,
        klass=Authenticator,
        allow_none=False,
        config=True,
        help="""
        The authenticator class to use for authentication.
        Default is DummyAuthenticator, which does not require a password by default.
        You can set this to your own authenticator class, which should inherit from Authenticator.
        """,
    )

    authenticator = Instance(klass=Authenticator)

    # TODO make configurable
    oauth_token_expires_in = int(1 * 24 * 3600)

    load_roles = Dict(
        List(),
        help="""
        Dict of `'<lecture-code>': List[{'members': List[Union[str, Dict[str, Optional[str]]]], 'role': str}]` entries to load at startup.

        Each lecture code maps to a list of role definitions. Each role definition must include:
          - 'role': the role name (e.g. 'student', 'tutor', 'instructor')
          - 'members': either
              * a list of username strings (legacy format), e.g. ['alice', 'bob']
              OR
              * a list of user objects, each with:
                  - 'username' (required): the unique username
                  - 'display_name' (optional): the user's display name; if omitted, defaults to the username

        Example::

            c.GraderService.load_roles = {
                'lecture1': [
                    {
                    'members': ['student1', 'student2'],  # legacy simple list
                    'role': 'student'
                    },
                    {
                    'members': [
                        {'username': 'instructor1', 'display_name': 'Instructor 1'},
                        {'username': 'instructor2'}
                    ],
                    'role': 'instructor'
                    }
                ],
            }


        """,
    ).tag(config=True)

    oauth_clients = List(
        Dict(),
        default_value=[],
        help="""
        List of OAuth clients 
        `[{'client_id': '<client_id>', 'client_secret': '<client_secret>', 'redirect_uri': '<redirect_uri>'}]`
        to register for the provider.
        
        Example::
            
            c.GraderService.oauth_clients = [{
                'client_id': 'hub',
                'client_secret': 'hub',
                'redirect_uri': 'http://localhost:8080/hub/oauth_callback'
            }]
    """,
    ).tag(config=True)

    @default("authenticator")
    def _authenticator_default(self):
        return self.authenticator_class(parent=self)

    @validate("config_file")
    def _validate_config_file(self, proposal):
        if not os.path.isfile(proposal.value) and not self.generate_config:
            print(
                "ERROR: Failed to find specified config file: {}".format(proposal.value),
                file=sys.stderr,
            )
            sys.exit(1)
        return proposal.value

    flags = {
        "debug": (
            {"Application": {"log_level": logging.DEBUG}},
            "Set log-level to debug, for the most verbose logging.",
        ),
        "show-config": (
            {"Application": {"show_config": True}},
            "Show the application's configuration (human-readable format)",
        ),
        "show-config-json": (
            {"Application": {"show_config_json": True}},
            "Show the application's configuration (json format)",
        ),
        "generate-config": (
            {"GraderService": {"generate_config": True}},
            "generate default config file",
        ),
    }

    aliases = {
        "log-level": "Application.log_level",
        "f": "GraderService.config_file",
        "config": "GraderService.config_file",
    }

    log_level = Enum(
        [
            0,
            10,
            20,
            30,
            40,
            50,
            "CRITICAL",
            "FATAL",
            "ERROR",
            "WARNING",
            "WARN",
            "INFO",
            "DEBUG",
            "NOTSET",
        ],
        "INFO",
    ).tag(config=True)

    def setup_loggers(self, log_level: str):  # pragma: no cover
        """Handles application, Tornado, and
        SQLAlchemy logging configuration."""
        stream_handler = logging.StreamHandler
        root_logger = logging.getLogger()
        root_logger.setLevel(log_level)
        fmt = "%(color)s%(levelname)-8s %(asctime)s %(module)-13s |%(end_color)s %(message)s"
        formatter = tornado.log.LogFormatter(fmt=fmt, color=True, datefmt=None)

        for log in ("access", "application", "general"):
            logger = logging.getLogger("tornado.{}".format(log))
            if len(logger.handlers) > 0:
                logger.removeHandler(logger.handlers[0])
            logger.setLevel(log_level)
            handler = stream_handler(stream=sys.stdout)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        sql_logger = logging.getLogger("sqlalchemy")
        sql_logger.propagate = False
        sql_logger.setLevel("WARN")
        sql_handler = stream_handler(stream=sys.stdout)
        sql_handler.setLevel("WARN")
        sql_handler.setFormatter(formatter)
        sql_logger.addHandler(sql_handler)

        oauth_log = logging.getLogger("oauthlib")
        oauth_handler = stream_handler(stream=sys.stdout)
        oauth_handler.setFormatter(formatter)
        oauth_log.setLevel(log_level)
        oauth_log.addHandler(oauth_handler)

        traitlet_logger = traitlets_log.get_logger()
        traitlet_logger.removeHandler(traitlet_logger.handlers[0])
        traitlet_logger.setLevel(log_level)
        traitlets_handler = stream_handler(stream=sys.stdout)
        traitlets_handler.setFormatter(formatter)
        traitlet_logger.addHandler(traitlets_handler)

    def write_config_file(self):
        self.log.info(f"Writing config file {os.path.abspath(self.config_file)}")
        config_file_dir = os.path.dirname(os.path.abspath(self.config_file))
        if not os.path.isdir(config_file_dir):
            self.exit(
                f"The directory to write the config file has to exist. {config_file_dir} not found"
            )
        if os.path.isfile(os.path.abspath(self.config_file)):
            self.exit(f"Config file {os.path.abspath(self.config_file)} already exists!")

        members = inspect.getmembers(
            sys.modules[__name__], lambda x: inspect.isclass(x) and issubclass(x, HasTraits)
        )
        config_classes = [x[1] for x in members]

        config_text = self.generate_config_file(classes=config_classes)
        if isinstance(config_text, bytes):
            config_text = config_text.decode("utf8")
        print("Generating config: %s" % self.config_file)
        with open(self.config_file, mode="w") as f:
            f.write(config_text)

    def initialize(self, argv, *args, **kwargs):
        self.log.info("Starting Initialization...")
        self.log.info("Loading config file...")
        super().initialize(*args, **kwargs)
        self.parse_command_line(argv)
        self.load_config_file(self.config_file)
        self.setup_loggers(self.log_level)

        self.session_maker = get_session_maker(self.db_url)
        self.init_roles()
        # use uvloop instead of default asyncio loop
        # asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        self._start_future = asyncio.Future()

        if sys.version_info.major < 3 or sys.version_info.minor < 9:
            msg = "Grader Service needs Python version 3.9 or above to run!"
            raise RuntimeError(msg)
        if shutil.which("git") is None:
            msg = "No git executable found! Git is necessary to run Grader Service!"
            raise RuntimeError(msg)

    def set_config(self):
        """Pass config to singletons."""
        RequestHandlerConfig.config = self.config
        LTISyncGrades.config = self.config
        CeleryApp.instance(config=self.config)

    async def cleanup(self):
        pass

    def init_oauth(self):
        engine = create_engine(self.db_url)
        session = sessionmaker(engine)
        self.oauth_provider = make_provider(
            session,
            url_prefix=url_path_join(self.base_url_path, "api/oauth2"),
            login_url=url_path_join(self.base_url_path, "login"),
            token_expires_in=self.oauth_token_expires_in,
        )
        for client in self.oauth_clients:
            self.oauth_provider.add_client(
                client["client_id"], client["client_secret"], client["redirect_uri"], ["identify"]
            )

    def init_roles(self):
        """Load predefined groups into the database"""
        with self.session_maker() as db:
            users_loaded = set()
            for lecture_code in self.load_roles.keys():
                role_list = self.load_roles[lecture_code]
                for role_dict in role_list:
                    role = role_dict.get("role")
                    users = role_dict.get("members", [])
                    lecture = db.query(Lecture).filter(Lecture.code == lecture_code).one_or_none()
                    # create lecture if no lecture with that name exists yet
                    # (code is set in create)
                    if lecture is None:
                        self.log.info(f"Adding new lecture with lecture_code {lecture_code}")
                        lecture = Lecture()
                        lecture.code = lecture_code
                        lecture.name = lecture_code
                        lecture.state = LectureState.active
                        lecture.deleted = DeleteState.active
                        db.add(lecture)
                    db.commit()

                    for user_entry in users:
                        if isinstance(user_entry, dict):
                            username = user_entry.get("username")
                            display_name = user_entry.get("display_name") or username
                        else:
                            # fallback for old format (just a username string)
                            username = user_entry
                            display_name = username

                        user = db.query(User).filter(User.name == username).one_or_none()
                        if user is None:
                            self.log.info(
                                f"Adding new user with username {username} "
                                f"and display name {display_name}"
                            )
                            user = User()
                            user.name = username
                            user.display_name = display_name
                            db.add(user)
                            db.commit()

                        # delete all roles of users the first time a new role is added for the user
                        if user.name not in users_loaded:
                            db.query(Role).filter(Role.user_id == user.id).delete()
                            users_loaded.add(user.name)

                        try:
                            db.add(Role(user_id=user.id, lectid=lecture.id, role=Scope[role]))
                        except KeyError:
                            self.log.error(f"Invalid role name: {role}")
                            raise ValueError(f"Invalid role name: {role}")
                    db.commit()

    async def start(self):
        self.log.info(f"Config File: {os.path.abspath(self.config_file)}")

        if self.generate_config:
            self.write_config_file()
            self.exit(0)

        self.log.info("Starting Grader Service...")
        self.io_loop = tornado.ioloop.IOLoop.current()

        self._setup_environment()

        self.init_oauth()

        # pass config
        self.set_config()

        handlers = HandlerPathRegistry.handler_list(self.base_url_path)
        self.log.info(handlers)
        # Add the handlers of the authenticator
        auth_handlers = self.authenticator.get_handlers(self.base_url_path)
        handlers.extend(auth_handlers)
        self.log.info(
            f"Registered authentication handlers for {self.authenticator.__class__.__name__}: "
            f"{[n for n, _ in auth_handlers]}"
        )

        oauth_provider_handlers = oauth_handlers.get_oauth_default_handlers(self.base_url_path)
        handlers.extend(oauth_provider_handlers)
        self.log.info(f"Registered OAuth handlers: {[n for n, _ in oauth_provider_handlers]}")

        # start the webserver
        self.http_server: HTTPServer = HTTPServer(
            GraderServer(
                grader_service_dir=self.grader_service_dir,
                base_url=self.base_url_path,
                authenticator=self.authenticator,
                handlers=handlers,
                oauth_provider=self.oauth_provider,
                cookie_secret=self.grader_cookie_secret,  # generate new cookie secret at startup
                config=self.config,
                session_maker=self.session_maker,
                parent=self,
                login_url=self.authenticator.login_url(self.base_url_path),
                logout_url=self.authenticator.logout_url(self.base_url_path),
                static_url_prefix=url_path_join(self.base_url_path, "/static/"),
                static_handler_class=CacheControlStaticFilesHandler,
                log_function=log_request,
            ),
            # ssl_options=ssl_context,
            max_buffer_size=self.max_buffer_size,
            max_body_size=self.max_body_size,
            xheaders=True,
        )
        self.log.info(f"Service directory - {self.grader_service_dir}")
        self.http_server.listen(
            self.service_port, address=self.service_host, reuse_port=self.reuse_port
        )

        for s in (signal.SIGTERM, signal.SIGINT):
            asyncio.get_event_loop().add_signal_handler(
                s, lambda: asyncio.ensure_future(self.shutdown_cancel_tasks(s))
            )

        self.log.info(f"Grader service running at {self.service_host}:{self.service_port}")

        # finish start
        self._start_future.set_result(None)

    def _setup_environment(self):
        if not os.path.exists(os.path.join(self.grader_service_dir, "git")):
            os.mkdir(os.path.join(self.grader_service_dir, "git"))
        # check if git config exits so that git commits don't fail
        if (
            subprocess.run(
                ["git", "config", "init.defaultBranch"], check=False, capture_output=True
            )
            .stdout.decode()
            .strip()
            != "main"
        ):
            raise RuntimeError("Git default branch has to be set to 'main'!")
        if (
            subprocess.run(["git", "config", "user.name"], check=False, capture_output=True)
            .stdout.decode()
            .strip()
            == ""
        ):
            raise RuntimeError("Git user.name has to be set!")
        if (
            subprocess.run(["git", "config", "user.email"], check=False, capture_output=True)
            .stdout.decode()
            .strip()
            == ""
        ):
            raise RuntimeError("Git user.email has to be set!")

    async def shutdown_cancel_tasks(self, sig):
        """Cancel all other tasks of the event loop and initiate cleanup"""
        self.log.critical("Received signal %s, initiating shutdown...", sig.name)

        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]

        if tasks:
            self.log.debug("Cancelling pending tasks")
            [t.cancel() for t in tasks]

            try:
                await asyncio.wait(tasks)
            except asyncio.CancelledError:
                self.log.debug("Caught Task CancelledError. Ignoring")
            except StopAsyncIteration:
                msg = "Caught StopAsyncIteration Exception"
                self.log.error(msg, exc_info=True)

            tasks = [t for t in asyncio.all_tasks()]
            for t in tasks:
                self.log.debug("Task status: %s", t)
        await self.cleanup()
        asyncio.get_event_loop().stop()

    async def launch_instance_async(self, argv=None):
        try:
            self.initialize(argv)
            await self.start()
        except Exception as e:
            self.log.exception(e)
            self.exit(1)

    @classmethod
    def launch_instance(cls, argv=None):
        self = cls.instance()
        loop = tornado.ioloop.IOLoop.current()
        task = asyncio.ensure_future(self.launch_instance_async(argv))
        try:
            loop.start()
        except KeyboardInterrupt:
            print("\nInterrupted")
        finally:
            if task.done():
                # re-raise exceptions in launch_instance_async
                task.result()
            loop.stop()

    @validate("grader_service_dir")
    def _validate_service_dir(self, proposal):
        path: str = proposal["value"]
        if not os.path.isabs(path):
            raise TraitError("The path is not absolute")
        if not os.path.isdir(path):
            os.mkdir(path, mode=0o700)
        return path

    @observe("grader_service_dir")
    def _observe_service_dir(self, change):
        path = change["new"]
        git_path = os.path.join(path, "git")
        if not os.path.isdir(git_path):
            os.mkdir(git_path, mode=0o700)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
from datetime import date, datetime, timezone
from typing import Any, Set, Union

from sqlalchemy import DECIMAL, Column, DateTime, Enum, ForeignKey, Integer, String, Text
from sqlalchemy.orm import relationship

from grader_service.api.models import assignment
from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.orm.base import Base, DeleteState, Serializable


def get_utc_time():
    return datetime.now(tz=timezone.utc)


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


class Assignment(Base, Serializable):
    __tablename__ = "assignment"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), nullable=False)
    """Name of the assignment"""
    lectid = Column(Integer, ForeignKey("lecture.id", ondelete="CASCADE"))
    points = Column(DECIMAL(10, 3), nullable=True)
    status = Column(Enum("created", "pushed", "released", "complete"), default="created")
    deleted = Column(Enum(DeleteState), nullable=False, unique=False)
    properties = Column(Text, nullable=True, unique=False)
    created_at = Column(DateTime, default=get_utc_time, nullable=False)
    updated_at = Column(DateTime, default=get_utc_time, onupdate=get_utc_time, nullable=False)
    _settings = Column("settings", Text, server_default="", nullable=False)

    lecture = relationship("Lecture", back_populates="assignments")
    submissions = relationship(
        "Submission",
        back_populates="assignment",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    @property
    def settings(self) -> AssignmentSettings:
        if self._settings is None:
            return AssignmentSettings()
        return AssignmentSettings.from_dict(json.loads(self._settings))

    @settings.setter
    def settings(self, settings: Union[AssignmentSettings, dict]):
        if isinstance(settings, dict):
            self._settings = json.dumps(settings, default=json_serial)
            return settings
        self._settings = json.dumps(settings.to_dict(), default=json_serial)
        return settings

    def update_settings(self, **kwargs: Any):
        # Update specific fields of the AssignmentSettings object
        settings = self.settings  # Get the current AssignmentSettings object
        for key, value in kwargs.items():
            if key not in AssignmentSettings.openapi_types.keys():
                raise RuntimeError(f"provided key '{key}' is not valid for assignment settings")
            if hasattr(settings, key):  # Ensure the attribute exists on AssignmentSettings
                setattr(settings, key, value)
        self.settings = settings  # Save the updated object back

    def get_whitelist_patterns(self) -> Set[str]:
        """
        Combines all whitelist patterns into a single set.
        """
        base_filter = ["*.ipynb"]
        extra_files = json.loads(self.properties).get("extra_files", [])
        allowed_file_patterns = self.settings.allowed_files

        return set(base_filter + extra_files + allowed_file_patterns)

    @property
    def model(self) -> assignment.Assignment:
        assignment_model = assignment.Assignment(
            id=self.id,
            name=self.name,
            status=self.status,
            points=self.points,
            settings=self.settings,
        )
        return assignment_model
